package v3d.editor;

import java.io.*;

// a line segment defined with two 3d vectors
class LineSegment {
	public Vector3D start;
	public Vector3D end;

	LineSegment() {
		start = new Vector3D();
		end = new Vector3D();
	}
}

// Camera class
// - rotate the camera
// - move the camera
// - change the fov
// - get a ray from 2-d coordinates on the viewplane
// - serialize/deserialize camera parameters
// - convert to VRML
class Camera
{
	public Matrix4x4 viewMatrix;
	public Matrix4x4 projectionMatrix;
	public float viewWidth;
	public float viewHeight;
	public float fov;
	public float viewXOffset;
	public float viewYOffset;
	public float minZ;
	public float maxZ;

	Camera() {
		viewMatrix = new Matrix4x4();
		projectionMatrix = new Matrix4x4();
	}

	public void setBase(Vector3D pos, Vector3D lookAt, Vector3D up) {
		viewMatrix.makeLookAt(pos,lookAt,up);
		viewMatrix.inverseFast();
	}

	public void setView(float width, float height, float fov, float minZ, float maxZ) {
		this.viewWidth = width;
		this.viewHeight = height;
		this.fov = fov  * (float)(2*Math.PI/360);
		this.minZ = minZ;
		this.maxZ = maxZ;
		projectionMatrix.makeProjection(this.viewWidth,this.viewHeight,this.fov,this.minZ,this.maxZ);
	}

	public void setFov(float fov) {
		// round to one decimal
		fov = (float)((int)(10 * fov))/10.0f;

		this.fov = fov * (float)(2*Math.PI/360);

		float minFov = (float)(2*Math.PI*5/360);
		float maxFov = (float)(2*Math.PI*160/360);
		if (this.fov < minFov)
			this.fov = minFov;
		else if (this.fov > maxFov)
			this.fov = maxFov;

		projectionMatrix.makeProjection(this.viewWidth,this.viewHeight,this.fov,this.minZ,this.maxZ);
	}

	public void setViewOffset(float xOffset, float yOffset) {
		this.viewXOffset = xOffset;
		this.viewYOffset = yOffset;
	}

	public void zoom(float fovChange)
	{
		float newFov = getFov() + fovChange;
		setFov(newFov);
	}

	public float getFov()
	{
		// round to one decimal
		float roundFov = (float)((int)(10 * fov * (float)(360/(2*Math.PI))))/10.0f;

		return roundFov;
	}

	public void move(Vector3D dir) {
		viewMatrix.inverseFast();
		viewMatrix.translate(dir);
		viewMatrix.inverseFast();
	}

	public void pan(float angle) {
		Matrix4x4 tmat = new Matrix4x4();
		tmat.makeRotationY(angle);

		viewMatrix.inverseFast();

		Vector3D translation = new Vector3D(viewMatrix.values[3][0],
								viewMatrix.values[3][1],
								viewMatrix.values[3][2]);
		viewMatrix.values[3][0] = 0.0f;
		viewMatrix.values[3][1] = 0.0f;
		viewMatrix.values[3][2] = 0.0f;

		viewMatrix = tmat.multiply(viewMatrix);
		viewMatrix.normalize();

		viewMatrix.values[3][0] = translation.x;
		viewMatrix.values[3][1] = translation.y;
		viewMatrix.values[3][2] = translation.z;

		viewMatrix.inverseFast();
	}

	public void tilt(float angle) {
		Matrix4x4 tmat = new Matrix4x4();
		tmat.makeRotationX(angle);

		viewMatrix.inverseFast();

		Vector3D translation = new Vector3D(viewMatrix.values[3][0],
								viewMatrix.values[3][1],
								viewMatrix.values[3][2]);
		viewMatrix.values[3][0] = 0.0f;
		viewMatrix.values[3][1] = 0.0f;
		viewMatrix.values[3][2] = 0.0f;

		viewMatrix = tmat.multiply(viewMatrix);
		viewMatrix.normalize();

		viewMatrix.values[3][0] = translation.x;
		viewMatrix.values[3][1] = translation.y;
		viewMatrix.values[3][2] = translation.z;

		viewMatrix.inverseFast();
	}

	public void roll(float angle) {
		Matrix4x4 tmat = new Matrix4x4();
		tmat.makeRotationZ(angle);

		viewMatrix.inverseFast();

		Vector3D translation = new Vector3D(viewMatrix.values[3][0],
								viewMatrix.values[3][1],
								viewMatrix.values[3][2]);
		viewMatrix.values[3][0] = 0.0f;
		viewMatrix.values[3][1] = 0.0f;
		viewMatrix.values[3][2] = 0.0f;

		viewMatrix = tmat.multiply(viewMatrix);
		viewMatrix.normalize();

		viewMatrix.values[3][0] = translation.x;
		viewMatrix.values[3][1] = translation.y;
		viewMatrix.values[3][2] = translation.z;

		viewMatrix.inverseFast();
	}

	public void rotateX(float angle) {
		Matrix4x4 tmat = new Matrix4x4();
		tmat.makeRotationX(angle);

		Vector3D translation = new Vector3D(viewMatrix.values[3][0],
								viewMatrix.values[3][1],
								viewMatrix.values[3][2]);
		viewMatrix.values[3][0] = 0.0f;
		viewMatrix.values[3][1] = 0.0f;
		viewMatrix.values[3][2] = 0.0f;

		viewMatrix.inverseFast();

		viewMatrix = tmat.multiply(viewMatrix);
		viewMatrix.normalize();

		viewMatrix.inverseFast();

		viewMatrix.values[3][0] = translation.x;
		viewMatrix.values[3][1] = translation.y;
		viewMatrix.values[3][2] = translation.z;
	}

	public void rotateY(float angle) {
		Matrix4x4 tmat = new Matrix4x4();
		tmat.makeRotationY(angle);

		Vector3D translation = new Vector3D(viewMatrix.values[3][0],
								viewMatrix.values[3][1],
								viewMatrix.values[3][2]);
		viewMatrix.values[3][0] = 0.0f;
		viewMatrix.values[3][1] = 0.0f;
		viewMatrix.values[3][2] = 0.0f;

		viewMatrix.inverseFast();

		viewMatrix = tmat.multiply(viewMatrix);
		viewMatrix.normalize();

		viewMatrix.inverseFast();

		viewMatrix.values[3][0] = translation.x;
		viewMatrix.values[3][1] = translation.y;
		viewMatrix.values[3][2] = translation.z;
	}

	public float getProjectedZPlane(float z) {
		return z * projectionMatrix.values[2][2] + projectionMatrix.values[3][2];
	}

	// returns a line segment in 3d space starting from near clip plane and
	// ending at far clip plane, where the ends of the segment are given as
	// 2d coordinates in the viewplane.
	public LineSegment getRay(int x, int y) {
		LineSegment seg = new LineSegment();
		float horizK = (float)((1.0f + Math.cos(fov))/Math.sin(fov));
		float vertK = viewWidth * horizK / viewHeight;
		float halfWidth = viewWidth / 2;
		float halfHeight = viewHeight / 2;

		float xp = ((float)x - halfWidth - viewXOffset)/(halfWidth * horizK);
		float yp = -((float)y - halfHeight - viewYOffset)/(halfHeight * vertK);

		Vector3D tp = new Vector3D();
		Matrix4x4 tmat = new Matrix4x4();

		tmat.copy(viewMatrix);
		tmat.inverseFast();

		tp.x = xp * minZ;
		tp.y = yp * minZ;
		tp.z = minZ;

		tmat.transform(tp,seg.start);

		tp.x = xp * maxZ;
		tp.y = yp * maxZ;
		tp.z = maxZ;

		tmat.transform(tp,seg.end);

		return seg;
	}

	// Camera is defined with a line starting with word "cam". The values on the line
	// are position of the camera (three floats), look at position (three floats),
	// up vector (three floats), fov (one float), near clip plane (one float) and far
	// clip plane (one float).

	// Creates a string of the camera
	public String serialize() {
		String s = new String();

		Matrix4x4 tMatrix = new Matrix4x4();
		tMatrix.copy(viewMatrix);
		tMatrix.inverseFast();

		Vector3D posVec = new Vector3D();
		tMatrix.transform(new Vector3D(0,0,0), posVec);

		Vector3D lookAtVec = new Vector3D();
		tMatrix.transform(new Vector3D(0,0,1), lookAtVec);

		Vector3D upVec = new Vector3D();
		tMatrix.transform(new Vector3D(0,1,0), upVec);

		upVec.subtractFromThis(posVec);

		s =	"cam " +
			posVec.x + " " +
			posVec.y + " " +
			posVec.z + " " +
			lookAtVec.x + " " +
			lookAtVec.y + " " +
			lookAtVec.z + " " +
			upVec.x + " " +
			upVec.y + " " +
			upVec.z + " " +
			fov * (float)(360/(2*Math.PI))+ " " +
			minZ + " " +
			maxZ + "\n";

		return s;
	}

	// Tries to read a camera from the stream. Returns
	// true if the token was correct and values were read,
	// otherwise false.
	public boolean deserialize(StreamTokenizer st) throws IOException {
		if ("cam".equals(st.sval)) {
			Vector3D posVec, lookAtVec, upVec;
			posVec = new Vector3D();
			lookAtVec = new Vector3D();
			upVec = new Vector3D();

			if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
				posVec.x = (float)st.nval;
			}
			if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
				posVec.y = (float)st.nval;
			}
			if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
				posVec.z = (float)st.nval;
			}
			if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
				lookAtVec.x = (float)st.nval;
			}
			if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
				lookAtVec.y = (float)st.nval;
			}
			if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
				lookAtVec.z = (float)st.nval;
			}
			if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
				upVec.x = (float)st.nval;
			}
			if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
				upVec.y = (float)st.nval;
			}
			if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
				upVec.z = (float)st.nval;
			}

			setBase(posVec,lookAtVec,upVec);

			if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
				this.fov = (float)st.nval * (float)(2*Math.PI/360);
			}
			if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
				this.minZ = (float)st.nval;
			}
			if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
				this.maxZ = (float)st.nval;
			}

			projectionMatrix.makeProjection(this.viewWidth,this.viewHeight,this.fov,this.minZ,this.maxZ);

			while (st.ttype != StreamTokenizer.TT_EOL &&
				st.ttype != StreamTokenizer.TT_EOF)
				st.nextToken();

			return true;
		}
		else {
			return false;
		}
	}

	public String getVRML() {
		String s = new String();

		Matrix4x4 tMatrix = new Matrix4x4();
		tMatrix.copy(viewMatrix);
		tMatrix.inverseFast();

		// take translation
		Vector3D translation = new Vector3D(tMatrix.values[3][0],tMatrix.values[3][1],tMatrix.values[3][2]);
		tMatrix.values[3][0] = 0;
		tMatrix.values[3][1] = 0;
		tMatrix.values[3][2] = 0;

		Vector3D xAxis = new Vector3D();
		Vector3D yAxis = new Vector3D();

		tMatrix.transform(new Vector3D(1,0,0), xAxis);
		tMatrix.transform(new Vector3D(0,1,0), yAxis);

		// take rotation around z
		// get the angle between the vector and coordinate axis
		float zAngle = getAngle(xAxis.x,xAxis.y) + (float)Math.PI/2;

		// remove rotation around z from the coordinates
		// After this the x-axis of the tMatrix is aligned to the z-x-plane in world coordinates
		tMatrix.makeRotationZ(zAngle);

		tMatrix.transform(new Vector3D(xAxis), xAxis);
		tMatrix.transform(new Vector3D(yAxis), yAxis);

		// take the rotation around y
		float yAngle = -getAngle(xAxis.x,xAxis.z) - (float)Math.PI/2;
		tMatrix.makeRotationY(yAngle);

		tMatrix.transform(new Vector3D(yAxis), yAxis);

		// get the rotation around x-axis
		float xAngle = getAngle(yAxis.y,yAxis.z) + (float)Math.PI/2;

		tMatrix.makeRotationX(xAngle);

		s = "Transform {\n" + 
			"\ttranslation\t" + translation.getVRML() + "\n" +
			"\tchildren [\n" +
				"\t\tTransform {\n" + 
					"\t\t\trotation\t0 0 1 " + (-zAngle) + "\n" +
					"\t\t\tchildren [\n" +
						"\t\t\t\tTransform {\n" + 
							"\t\t\t\t\trotation\t0 1 0 " + yAngle + "\n" +
							"\t\t\t\t\tchildren [\n" +
								"\t\t\t\t\t\tViewpoint {\n" +
									"\t\t\t\t\t\t\tfieldOfView\t" + fov + "\n" +
									"\t\t\t\t\t\t\tjump\tTRUE\n" +
									"\t\t\t\t\t\t\torientation\t1 0 0 " + xAngle + "\n" +
									"\t\t\t\t\t\t\tposition\t0 0 0\n"+
									"\t\t\t\t\t\t\tdescription\t\"camera\"\n" +
								"\t\t\t\t\t\t}\n" +
							"\t\t\t\t\t]\n" +
						"\t\t\t\t}\n" +
					"\t\t\t]\n" +
				"\t\t}\n" +
			"\t]\n" +
		"}\n";
		
		return s;
	}

	float getAngle(float x, float y)
	{
		float a;

		if(y!=(0.0f))
			a=(float)Math.atan(x/y);
		else if(x!=(0.0f))
			a=((float)Math.PI/2)*(float)Math.abs(x)/x;
		else
			a=0.0f;

		if(y>0.0f)
			a+=(float)Math.PI;
		else if(y==0.0) {
			if (x>0.0)
				a+=(float)Math.PI;
		}
		if (a>(float)Math.PI)
			a-=(float)Math.PI*2;

		return a;
	}
}
