/* 
 * Editor.java
 * 
 * This is the editor for 3d6b project.
 * Written by Kiia Kallio in 2002.
 * 
 */

package v3d.editor;

import java.applet.Applet;
import java.lang.*;
import java.awt.*;
import java.awt.geom.*;
import java.awt.Graphics2D;
import java.net.*;
import java.net.URL;
import java.util.Vector;


/*
 * Editor class is the main class file for the editor applet. It acts as a
 * "glue" for display, UI and scene handling.
 */
public class Editor extends Applet implements Runnable {
	Thread engine = null;
	Panel view;
	MessagesPanel messages;
	UIPanel ui;
	FileIO io;
	UIControls controls;
	boolean isApplet = true;
	String currentEditor;
	RenderingSettings renderingSettings;

	public void init() {
		configRendering();

		setLayout(new BorderLayout());
		ui = new UIPanel(this);
		if (isApplet)
			io = new AppletFileIO(this);
		else
			io = new ApplicationFileIO(this);

		messages = new MessagesPanel(this);

		// view is a CardLayout for switching between scene view and message/history view
 		view = new Panel();
		view.setLayout(new CardLayout());
		view.add("UI",ui);
		view.add("Messages",messages);

		controls = new UIControls(this);
		add("Center", view);
		add("West", controls);
	}


	public static void main(String args[]) {
		ApplicationFrame f = new ApplicationFrame("3D6B Editor");
		Editor ed = new Editor();

		f.show();
		Insets i = f.insets();
		f.hide();

//		System.out.println("top: "+i.top+" left: "+i.left+" bottom: "+i.bottom+" right: "+i.right);

		f.resize(800 + i.right + i.left, 400 + i.bottom + i.top);
		f.setResizable(false);

		f.add("Center", ed);
		ed.isApplet = false;
		ed.init();

		f.show();

		ed.start();
	}

	// start the applet
	public void start() {
		ui.start();
		io.start();
		notifySceneLoaded();
		controls.requestFocus();
	}

	// stop the applet
 	public void stop() {
		ui.stop();
	}

	public void run()
	{

	}

	// print to the console
	private void print(String s) {
		System.out.println(s); 		// console
	}

	public Scene getScene() {
		return ui.getScene();
	}

	public void setUserMsg(String msg) {
		messages.setMessage(msg);
	}

	public String getUserMsg() {
		return messages.getMessage();
	}

	public void showUI() {
		((CardLayout)view.getLayout()).show(view,"UI");
	}

	public void showMessages() {
		((CardLayout)view.getLayout()).show(view,"Messages");
	}
	
	public byte[] getThumbnailTGA(int width, int height, int scale) {
		return ui.getThumbnailTGA(width,height,scale);
	}

	public FileIO getFileIO() {
		return io;
	}

	public UIControls getControls() {
		return controls;
	}

	public UIPanel getUI() {
		return ui;
	}

	public void setCurrentEditor(String editingInfo) {
		currentEditor = editingInfo;
	}

	public void notifySceneLoaded() {
		Scene scene = ui.getScene();
		scene.setEditor(currentEditor);
		controls.notifyFOVChange(scene.cam);
		controls.notifyGridChange(scene.grid);
		messages.setHistory(scene.history);
	}

	public RenderingSettings getRenderingSettings() {
		return renderingSettings;
	}

	void configRendering() {
		boolean supportsJava2D = false;
		try {
			// java.awt.geom.Point2D only exists with Java2D enabled VMs
			Class.forName("java.awt.geom.Point2D");
			supportsJava2D = true;
		} catch (Exception e) { }

/*		if (supportsJava2D)
			System.out.println("Java2D");
		else
			System.out.println("No Java2D");*/

		renderingSettings = null;
		try {
			if (supportsJava2D) {
				renderingSettings = (RenderingSettings)Class.forName("v3d.editor.Java2DRenderingSettings").newInstance();
			}
			else {
				renderingSettings = (RenderingSettings)Class.forName("v3d.editor.RenderingSettings").newInstance();
			}
		}
		catch (Exception e) {}
	}
}

class ApplicationFrame extends Frame {
	ApplicationFrame(String title) {
		setTitle(title);
	}

	public boolean handleEvent(Event e) {
//		System.out.println(e.toString());
		switch (e.id) {
			case Event.WINDOW_DESTROY:
				System.exit(0);
				return true;
			default:
				return false;
		}
	}
}

/*
 * UIPanel is a class that handles the display of the scene and the grid and
 * sends mouse actions that happen on the drawing area to objects that are
 * currently manipulated, wheter it is the scene, the grid or the camera.
 */
class UIPanel extends Panel implements Runnable {
	public static final int DRAW = 0;
	public static final int GRID = 1;
	public static final int CAMERA = 2;
	public static final int FILE = 3;
	public static final int DRAW_FREEHAND = 4;
	public static final int DRAW_LINE = 5;
	public static final int DRAW_ERASER = 6;
	public static final int GRID_MOVE = 7;
	public static final int GRID_PUSHROLL = 8;
	public static final int GRID_ROTATE = 9;
	public static final int GRID_BEND = 10;
	public static final int CAMERA_MOVE = 11;
	public static final int CAMERA_PANTILT = 12;
	public static final int CAMERA_ROTATE = 13;
	public static final int CAMERA_DOLLYROLL = 14;
	public static final int CAMERA_ZOOM = 15;
	int mode = DRAW;
	int drawMode = DRAW_FREEHAND;
	int gridMode = GRID_MOVE;
	int cameraMode = CAMERA_MOVE;
	int currentDrawingColor = 0;
	float eraserSize = 10;
	boolean historyView = false;
	float keySpeed = 10.0f;

	Thread engine = null;

	Image imageDoubleBuffer;
	Graphics gDoubleBuffer = null;

	int prevX;
	int prevY;

	boolean drawState;
	boolean prevHit;
	Vector3D prevPnt;
	int currentLine;

	Scene scene;

	Editor editor;

	public UIPanel(Editor ed) {
		setBackground(Color.white);
		this.prevPnt = new Vector3D();
		this.prevHit = false;
		this.drawState = false;
		this.editor = ed;
	}

	public void setUIMode(int mode) {
		switch (mode) {
			case DRAW:
			case GRID:
			case CAMERA:
			case FILE:
				this.mode = mode;
//				System.out.println("Mode: "+this.mode);
				break;
			default:
				throw new IllegalArgumentException();
		}
	}

	public void setDrawMode(int drawMode) {
		switch (drawMode) {
			case DRAW_FREEHAND:
			case DRAW_LINE:
			case DRAW_ERASER:
				this.drawMode = drawMode;
//				System.out.println("DrawMode: "+this.drawMode);
				break;
			default:
				throw new IllegalArgumentException();
		}
	}

	public void setGridMode(int gridMode) {
		switch (gridMode) {
			case GRID_MOVE:
			case GRID_PUSHROLL:
			case GRID_ROTATE:
			case GRID_BEND:
				this.gridMode = gridMode;
//				System.out.println("GridMode: "+this.gridMode);
				break;
			default:
				throw new IllegalArgumentException();
		}
	}

	public void setGridVisibility(boolean visible) {
		if (visible)
			scene.showGrid();
		else
			scene.hideGrid();
	}

	public void setGridStepping(boolean stepping) {
		scene.grid.setStepping(stepping);
	}

	public void setGridStepSize(float stepSize) {
		scene.grid.setStepSize(stepSize);
	}

	public void setGridSnapping(boolean snapping) {
		scene.grid.setSnapping(snapping);
	}

	public void setGridSize(float gridSize) {
		scene.grid.setSize(gridSize);
	}

	public void resetGrid() {
		scene.grid.reset();
		editor.getControls().notifyGridChange(scene.grid);
	}

	public void setCameraMode(int cameraMode) {
		switch (cameraMode) {
			case CAMERA_MOVE:
			case CAMERA_PANTILT:
			case CAMERA_ROTATE:
			case CAMERA_DOLLYROLL:
			case CAMERA_ZOOM:
				this.cameraMode = cameraMode;
//				System.out.println("CameraMode: "+this.cameraMode);
				break;
			default:
				throw new IllegalArgumentException();
		}
	}

	public void setCurrentDrawingColor(int color)
	{
// 		System.out.println("DrawingColor: "+color);
		currentDrawingColor = color;
	}

	public void setCurrentHistoryColor(int color)
	{
// 		System.out.println("HistoryColor: "+color);
		scene.history.setCurrentColor(color);
	}

	public int getCurrentDrawingColor()
	{
		return currentDrawingColor;
	}

	public int getCurrentHistoryColor()
	{
		return scene.history.getCurrentColor();
	}

	public int getUIMode() {
		return this.mode;
	}

	public Scene getScene() {
		return this.scene;
	}

	public void setHistoryView(boolean viewState)
	{
// 		System.out.println("HistoryView: "+viewState);
		historyView = viewState;
	}

	public boolean isHistoryView()
	{
		return historyView;
	}

	public boolean handleEvent(Event e) {
		int diffX, diffY;
		boolean hit;
		Vector3D point;
		LineSegment ray;
		switch (e.id) {
			case Event.MOUSE_DOWN:
				switch (mode) {
					case DRAW:
						scene.startCommand();
						ray = scene.cam.getRay(e.x,e.y);
						point = new Vector3D();
						hit = scene.grid.getIntersection(ray,point);
						if (hit) {
							if (drawMode == DRAW_ERASER) {
								scene.lineset.erase(point,eraserSize);
							}
						}
						prevHit = false;
						drawState = hit;
						prevPnt.copy(point);
						break;
					case GRID:
						scene.startCommand();
						scene.grid.resetStepCumulation();
						break;
					case CAMERA:
						break;
					default:
						break;
				}
				prevX = e.x;
				prevY = e.y;
				return true;
			case Event.MOUSE_UP:
				switch (mode) {
					case DRAW:
						ray = scene.cam.getRay(e.x,e.y);
						point = new Vector3D();
						hit = scene.grid.getIntersection(ray,point);
						if (hit) {
							switch (drawMode) {
								case DRAW_FREEHAND:
									if (drawState) {
										if (!prevPnt.isEqual(point)) {
											if (!prevHit) {
												scene.lineset.line(prevPnt,point,currentDrawingColor);
											}
											else {
												scene.lineset.lineFromPreviousEnd(point,currentDrawingColor);
											}
										}
									}
									break;
								case DRAW_LINE:
									if (drawState) {
										if (!prevPnt.isEqual(point)) {
											if (!prevHit) {
												scene.lineset.line(prevPnt,point,currentDrawingColor);
											}
											else {
												scene.lineset.changeLineEnd(currentLine,point);
											}
										}
									}
									break;
								case DRAW_ERASER:
									break;
								default:
									break;
							}
						}
						scene.endCommand();
						break;
					case GRID:
						scene.grid.resetStepCumulation();
						scene.endCommand();
						break;
					case CAMERA:
						break;
					default:
						break;
				}
				drawState = false;
				prevHit = false;
				return true;
			case Event.MOUSE_DRAG:
				diffX = (e.x - prevX);
				diffY = (e.y - prevY);
				switch (mode) {
					case DRAW:
						ray = scene.cam.getRay(e.x,e.y);
						point = new Vector3D();
						hit = scene.grid.getIntersection(ray,point);
						if (hit) {
							switch (drawMode) {
								case DRAW_FREEHAND:
									if (drawState) {
										if (!prevPnt.isEqual(point)) {
											if (!prevHit) {
													scene.lineset.line(prevPnt,point,currentDrawingColor);
											}
											else {
												if (!prevPnt.isEqual(point))
													scene.lineset.lineFromPreviousEnd(point,currentDrawingColor);
											}
											prevHit = true;
										}
									}
									else {
										drawState = true;
									}
									prevPnt.copy(point);
									break;
								case DRAW_LINE:
									if (drawState) {
										if (!prevPnt.isEqual(point)) {
											if (!prevHit) {
												currentLine = scene.lineset.line(prevPnt,point,currentDrawingColor);
											}
											else {
												scene.lineset.changeLineEnd(currentLine,point);
											}
											prevHit = true;
										}	
									}
									else {
										drawState = true;
									}
									prevPnt.copy(point);
									break;
								case DRAW_ERASER:
									scene.lineset.erase(point,eraserSize);
									break;
								default:
									break;
							}
						}
						break;
					case GRID:
						switch (gridMode) {
				  			case GRID_MOVE:
								scene.grid.move(new Vector3D((float)diffX,
												(float)-diffY,
												0.0f),scene.cam,false);
								break;
				  			case GRID_PUSHROLL:
								if (Math.abs(diffX) > Math.abs(diffY))
									scene.grid.rotateZ((float)diffX,scene.cam,false);
								else
									scene.grid.move(new Vector3D(0.0f,
												0.0f,
												(float)-diffY),scene.cam,false);
								break;
				  			case GRID_ROTATE:
								if (Math.abs(diffX) > Math.abs(diffY))
									scene.grid.rotateY((float)-diffX,scene.cam,false);
								else
									scene.grid.rotateX((float)-diffY,scene.cam,false);
								break;
				  			case GRID_BEND:
								if (Math.abs(diffX) > Math.abs(diffY))
									scene.grid.bendZ((float)-diffX,false);
								else
									scene.grid.bendX((float)-diffY,false);
								break;
							default:
								break;
						}
						break;
					case CAMERA:
						switch (cameraMode) {
							case CAMERA_MOVE:
								scene.cam.move(new Vector3D((float)-diffX*0.5f,
												(float)diffY*0.5f,
												0.0f));
								break;
							case CAMERA_PANTILT:
								if (Math.abs(diffX) > Math.abs(diffY))
									scene.cam.pan((float)-diffX*0.001f);
								else
									scene.cam.tilt((float)-diffY*0.001f);
								break;
							case CAMERA_ROTATE:
								if (Math.abs(diffX) > Math.abs(diffY))
									scene.cam.rotateY((float)diffX*0.005f);
								else
									scene.cam.rotateX((float)diffY*0.005f);
								break;
							case CAMERA_DOLLYROLL:
								if (Math.abs(diffX) > Math.abs(diffY))
									scene.cam.roll((float)-diffX*0.005f);
								else
									scene.cam.move(new Vector3D(0.0f,
												0.0f,
												(float)diffY));
								break;
							case CAMERA_ZOOM:
								scene.cam.zoom((float)diffY*0.1f);
								editor.getControls().notifyFOVChange(scene.cam);
								break;
							default:
								break;
						}
						break;
					default:
						break;
				}
				prevX = e.x;
				prevY = e.y;
				return true;
			case Event.KEY_PRESS:
				switch (e.key) {
					// q = 113 w = 119 e = 101
					// a = 97  s = 115 d = 100
					case 97: //a
						scene.startCommand();
						scene.grid.move(new Vector3D(0.0f,keySpeed,0.0f),true);
						scene.endCommand();
						break;
					case 100: //d
						scene.startCommand();
						scene.grid.move(new Vector3D(0.0f,-keySpeed,0.0f),true);
						scene.endCommand();
						break;
					case 119: //w
						scene.startCommand();
						scene.grid.move(new Vector3D(keySpeed,0.0f,0.0f),true);
						scene.endCommand();
						break;
					case 115: //s
						scene.startCommand();
						scene.grid.move(new Vector3D(-keySpeed,0.0f,0.0f),true);
						scene.endCommand();
						break;
					case 113: //q
						scene.startCommand();
						scene.grid.move(new Vector3D(0.0f,0.0f,keySpeed),true);
						scene.endCommand();
						break;
					case 101: //e
						scene.startCommand();
						scene.grid.move(new Vector3D(0.0f,0.0f,-keySpeed),true);
						scene.endCommand();
						break;
					// r = 114 t = 116 y = 121
					// f = 102 g = 103 h = 104
					case 102: //f
						scene.startCommand();
						scene.grid.rotateY(-keySpeed,true);
						scene.endCommand();
						break;
					case 104: //h
						scene.startCommand();
						scene.grid.rotateY(keySpeed,true);
						scene.endCommand();
						break;
					case 116: //t
						scene.startCommand();
						scene.grid.rotateZ(-keySpeed,true);
						scene.endCommand();
						break;
 					case 103: //g
						scene.startCommand();
						scene.grid.rotateZ(keySpeed,true);
						scene.endCommand();
						break;
					case 114: //r
						scene.startCommand();
						scene.grid.rotateX(keySpeed,true);
						scene.endCommand();
						break;
					case 121: //y
						scene.startCommand();
						scene.grid.rotateX(-keySpeed,true);
						scene.endCommand();
						break;
					// 1 = 49  2 = 50  3 = 51  4 = 52
					case 49: //1
						scene.startCommand();
						scene.grid.bendX(keySpeed,true);
						scene.endCommand();
						break;
					case 50: //2
						scene.startCommand();
						scene.grid.bendX(-keySpeed,true);
						scene.endCommand();
						break;
					case 51: //3
						scene.startCommand();
						scene.grid.bendZ(keySpeed,true);
						scene.endCommand();
						break;
					case 52: //4
						scene.startCommand();
						scene.grid.bendZ(-keySpeed,true);
						scene.endCommand();
						break;
				}
//				System.out.println(e.toString());
				return true;
			case Event.KEY_RELEASE:
				scene.grid.resetStepCumulation();
				return true;
			case Event.KEY_ACTION:
				return true;
			case Event.KEY_ACTION_RELEASE:
				return true;
			case Event.WINDOW_DESTROY:
				System.exit(0);
				return true;
			default:
				return false;
		}
	}

	public void setFov(float fov) {
		scene.cam.setFov(fov);
	}

	public void update(Graphics g) {    
		paint(g);
	}

	public void paint(Graphics g) {
		// clear the screen

		gDoubleBuffer.setColor(Color.white);
		gDoubleBuffer.fillRect(0, 0, size().width, size().height);

		// render the scene
		if (historyView)
			scene.renderHistory(gDoubleBuffer);
		else
			scene.render(gDoubleBuffer);

		// draw all the stuff from the double buffer to the screen
		g.drawImage(imageDoubleBuffer, 0, 0, null);
	}

	// when the UI panel gets the layout() call, it allocates the double
	// buffer according to its size and creates the scene.
	public void layout()
	{
		imageDoubleBuffer = createImage(size().width, size().height);
		gDoubleBuffer = imageDoubleBuffer.getGraphics();
		gDoubleBuffer.setColor(Color.white);
		gDoubleBuffer.fillRect(0, 0, size().width, size().height);

		setAntialias(true);

		scene = new Scene(size().width,size().height);
	}

	public void setAntialias(boolean state) {
		RenderingSettings rs = editor.getRenderingSettings();
		if (rs != null && gDoubleBuffer != null) {
			if (state)
				rs.enableAntialias(gDoubleBuffer);
			else
				rs.disableAntialias(gDoubleBuffer);
		}
	}

	// start the UI
	public void start() {
		if (engine == null) {
			engine = new Thread(this);
			engine.start();
		}
	}

	// stop the UI
 	public void stop() {
		if (engine != null && engine.isAlive()) {
			engine.stop();
		}
		engine = null; 
	}

	// run the UI thread
	public void run()
	{
		while (true) {
			try {
				repaint();
				System.gc();
				Thread.sleep(50);
			}
			catch (InterruptedException e) {
				stop();
			}
		}
	}

	public byte[] getThumbnailTGA(int width, int height, int scale) {
		int scaledWidth = width*scale;
		int scaledHeight = height*scale;

		Image img = createImage(scaledWidth,scaledHeight);
		Graphics gi = img.getGraphics();
		gi.setColor(Color.white);
		gi.fillRect(0, 0, scaledWidth, scaledHeight);

		int w = imageDoubleBuffer.getWidth(null);
		int h = imageDoubleBuffer.getHeight(null);

		if (w < 0) w = scaledWidth;
		if (h < 0) h = scaledHeight;

		gi.drawImage(imageDoubleBuffer,(scaledWidth - w)/2,(scaledHeight - h)/2,null);

		TGAEncoder encoder = new TGAEncoder(img, scale);
		return encoder.getTGA();
	}
}

/*
 * This is a base class for configuring rendering
 */
class RenderingSettings {
	RenderingSettings() { };
	public boolean isAntialiasAvailable() { return false; };
	public void enableAntialias(Graphics gr) { };
	public void disableAntialias(Graphics gr) { };
}

/*
 * This is a class that can be used for enabling Java2D options (if available)
 */
class Java2DRenderingSettings extends RenderingSettings {
	Java2DRenderingSettings() {};
	public boolean isAntialiasAvailable() { return true; };	
	public void enableAntialias(Graphics gr) {
		Graphics2D g = (Graphics2D)gr;
		RenderingHints hints = new RenderingHints(null); 
		hints.put(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON);
		g.addRenderingHints(hints);
	}
	public void disableAntialias(Graphics gr) {
		Graphics2D g = (Graphics2D)gr;
		RenderingHints hints = new RenderingHints(null); 
		hints.put(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_OFF);
		g.addRenderingHints(hints);
	}
}

/*
 * UIControls is a class that contains all the buttons and other such elements
 * of the interface. It acts as the message dispatcher between the interface elements
 * and UIPanel, Grid, Camera and Scene.
 */
class UIControls extends Panel3D {
	Editor editor;
	Panel optionsPanel;
	ColorPicker colorPicker;
	TextField gridStepSize;
	TextField gridSize;
	TextField cameraFOV;
	Checkbox gridStep;
	Checkbox gridSnap;
	Checkbox gridHide;

	public UIControls(Editor ed) {
		this.editor = ed;

		Color drawColor = new Color(181,205,222);
		Color gridColor; // = new Color(236,215,150);
		Color cameraColor; // = new Color(238,179,148);
		Color fileColor; // = new Color(210,178,205);

		Color lightDrawColor = new Color((181+255)/2,(205+255)/2,(222+255)/2);
		Color lightGridColor; // = new Color((236+255)/2,(215+255)/2,(150+255)/2);
		Color lightCameraColor; // = new Color((238+255)/2,(179+255)/2,(148+255)/2);
		Color lightFileColor; // = new Color((210+255)/2,(178+255)/2,(205+255)/2);

		Color bgColor; // = new Color(249,242,220);

		bgColor = drawColor;
		gridColor = drawColor;
		cameraColor = drawColor;
		fileColor = drawColor;
		lightGridColor = lightDrawColor;
		lightCameraColor = lightDrawColor;
		lightFileColor = lightDrawColor;

		setBackground(bgColor);

		// main mode switch panel
		Panel3D modePanel = new Panel3D();
		modePanel.setBackground(lightDrawColor);
		modePanel.setLayout(new FlowLayout(FlowLayout.CENTER,1,1));

		SwitchPanel modeSwitch = new SwitchPanel(this);
		modeSwitch.add("Draw",UIPanel.DRAW,drawColor);
		modeSwitch.add("Grid",UIPanel.GRID,gridColor);
		modeSwitch.add("Camera",UIPanel.CAMERA,cameraColor);
		modeSwitch.add("File",UIPanel.FILE,fileColor);

		modePanel.add(modeSwitch);

		// options panel is a CardLayout for switching options
		// in various drawing modes
		optionsPanel = new Panel();
		optionsPanel.setLayout(new CardLayout());

		// DRAW OPTIONS PANEL
		// ------------------
		Panel3D drawOptions = new Panel3D();
		drawOptions.setBackground(drawColor);
		GridBagLayout gridbag = new GridBagLayout();
		GridBagConstraints c = new GridBagConstraints();
		drawOptions.setLayout(gridbag);

		c.insets = new Insets(2,2,2,2);

		c.fill = GridBagConstraints.BOTH;
		c.anchor = GridBagConstraints.NORTH;
		c.gridwidth  = GridBagConstraints.REMAINDER;
		c.weightx = 0.0;
		c.weighty = 0.0;

		// first a switch for draw modes
		SwitchPanel drawSwitch = new SwitchPanel(this);
		drawSwitch.add("Freehand",UIPanel.DRAW_FREEHAND,lightDrawColor);
		drawSwitch.add("Line",UIPanel.DRAW_LINE,lightDrawColor);
		drawSwitch.add("Eraser",UIPanel.DRAW_ERASER,lightDrawColor);

		gridbag.setConstraints(drawSwitch, c);
		drawOptions.add(drawSwitch);

		// add a color picker
		colorPicker = new ColorPicker(this,80,21);
		gridbag.setConstraints(colorPicker, c);
		drawOptions.add(colorPicker);

		// and another panel containing the rest of the stuff
		Panel drawOptionsContainer = new Panel();
		drawOptionsContainer.setLayout(new FlowLayout(FlowLayout.CENTER,1,1));

		Panel drawOptionsLayout = new Panel();
		drawOptionsLayout.setLayout(new GridLayout(3,1,2,2));

		drawOptionsLayout.add(new Button("Undo"));
		drawOptionsLayout.add(new Button("Redo"));
		drawOptionsLayout.add(new Checkbox("History"));

		c.fill = GridBagConstraints.BOTH;
		c.anchor = GridBagConstraints.NORTH;
		c.gridheight = GridBagConstraints.REMAINDER;
		c.weightx = 1.0;
		c.weighty = 1.0;

		drawOptionsContainer.add(drawOptionsLayout);
		gridbag.setConstraints(drawOptionsContainer, c);
		drawOptions.add(drawOptionsContainer);

		optionsPanel.add("Draw",drawOptions);

		// GRID OPTIONS PANEL
		// ------------------
		Panel3D gridOptions = new Panel3D();
		gridOptions.setBackground(gridColor);
		gridbag = new GridBagLayout();
		c = new GridBagConstraints();
		gridOptions.setLayout(gridbag);

		c.insets = new Insets(2,2,2,2);

		c.fill = GridBagConstraints.BOTH;
		c.anchor = GridBagConstraints.NORTH;
		c.gridwidth  = GridBagConstraints.REMAINDER;
		c.weightx = 0.0;
		c.weighty = 0.0;

		// first a switch for grid editing tools
		SwitchPanel gridSwitch = new SwitchPanel(this);
		gridSwitch.add("Move",UIPanel.GRID_MOVE,lightGridColor);
		gridSwitch.add("Push/Roll",UIPanel.GRID_PUSHROLL,lightGridColor);
		gridSwitch.add("Rotate",UIPanel.GRID_ROTATE,lightGridColor);
		gridSwitch.add("Bend",UIPanel.GRID_BEND,lightGridColor);

		gridbag.setConstraints(gridSwitch,  c);
		gridOptions.add(gridSwitch);

		// and another panel containing the rest of the stuff
		Panel gridOptionsLayout = new Panel();
		gridOptionsLayout.setLayout(new FlowLayout(FlowLayout.CENTER,0,0));

		Panel tempPanel = new Panel();
		tempPanel.setLayout(new GridLayout(4,1,2,2));

		Panel p = new Panel();
		p.setLayout(new FlowLayout(FlowLayout.RIGHT,2,0));

		gridStep = new Checkbox("Step:");
		p.add(gridStep);
		gridStepSize = new TextField("100");
		p.add(gridStepSize);

		tempPanel.add(p);

		p = new Panel();
		p.setLayout(new FlowLayout(FlowLayout.RIGHT,2,0));

		p.add(new Label("Size:",Label.LEFT));
		gridSize = new TextField("100");
		p.add(gridSize);

		tempPanel.add(p);

		p = new Panel();
		p.setLayout(new FlowLayout(FlowLayout.CENTER,0,0));

		gridSnap = new Checkbox("Snap");
		p.add(gridSnap);

		gridHide = new Checkbox("Hide");
		p.add(gridHide);

		tempPanel.add(p);

		p = new Panel();
		p.setLayout(new FlowLayout(FlowLayout.CENTER,0,0));
		p.add(new Button("Reset"));
		tempPanel.add(p);

		gridOptionsLayout.add(tempPanel);

		c.fill = GridBagConstraints.BOTH;
		c.anchor = GridBagConstraints.NORTH;
		c.gridheight = GridBagConstraints.REMAINDER;
		c.weightx = 1.0;
		c.weighty = 1.0;

		gridbag.setConstraints(gridOptionsLayout,  c);
		gridOptions.add(gridOptionsLayout);

		optionsPanel.add("Grid",gridOptions);

		// CAMERA OPTIONS PANEL
		// --------------------
		Panel3D camOptions = new Panel3D();
		camOptions.setBackground(cameraColor);
		gridbag = new GridBagLayout();
		c = new GridBagConstraints();
		camOptions.setLayout(gridbag);

		c.insets = new Insets(2,2,2,2);

		c.fill = GridBagConstraints.BOTH;
		c.anchor = GridBagConstraints.CENTER;
		c.gridwidth  = GridBagConstraints.REMAINDER;
		c.weightx = 0.0;
		c.weighty = 0.0;

		// first a switch for camera editing tools
		SwitchPanel camSwitch = new SwitchPanel(this);
		camSwitch.add("Move",UIPanel.CAMERA_MOVE,lightCameraColor);
		camSwitch.add("Pan/Tilt",UIPanel.CAMERA_PANTILT,lightCameraColor);
		camSwitch.add("Rotate",UIPanel.CAMERA_ROTATE,lightCameraColor);
		camSwitch.add("Dolly/Roll",UIPanel.CAMERA_DOLLYROLL,lightCameraColor);
		camSwitch.add("Zoom",UIPanel.CAMERA_ZOOM,lightCameraColor);

		gridbag.setConstraints(camSwitch,  c);
		camOptions.add(camSwitch);

		// and another panel containing the rest of the stuff
		Panel camOptionsContainer = new Panel();
		camOptionsContainer.setLayout(new FlowLayout(FlowLayout.CENTER,0,0));

		boolean aaAvailable = false;
		RenderingSettings rs = editor.getRenderingSettings();
		if (rs != null)
			aaAvailable = rs.isAntialiasAvailable();

		Panel camOptionsLayout = new Panel();
		if (aaAvailable)
			camOptionsLayout.setLayout(new GridLayout(3,1,2,2));
		else
			camOptionsLayout.setLayout(new GridLayout(2,1,2,2));

		camOptionsLayout.add(new Label("FOV:",Label.CENTER));
		cameraFOV = new TextField("60.");
		camOptionsLayout.add(cameraFOV);

		if (aaAvailable) {
			Checkbox aa = new Checkbox("Antialias");
			aa.setState(true);
			camOptionsLayout.add(aa);
		}

		c.fill = GridBagConstraints.BOTH;
		c.anchor = GridBagConstraints.CENTER;
		c.gridheight = GridBagConstraints.REMAINDER;
		c.weightx = 1.0;
		c.weighty = 1.0;

		camOptionsContainer.add(camOptionsLayout);
		gridbag.setConstraints(camOptionsContainer,  c);
		camOptions.add(camOptionsContainer);

		optionsPanel.add("Camera",camOptions);

		// FILE OPTIONS PANEL
		// ------------------
		Panel3D fileOptions = new Panel3D();
		fileOptions.setBackground(fileColor);
		fileOptions.setLayout(new FlowLayout(FlowLayout.CENTER,1,1));

		Panel fileOptionsContainer = new Panel();

		gridbag = new GridBagLayout();
		c = new GridBagConstraints();
		fileOptionsContainer.setLayout(gridbag);

		c.insets = new Insets(2,2,2,2);

		c.fill = GridBagConstraints.BOTH;
		c.anchor = GridBagConstraints.CENTER;
		c.gridwidth  = GridBagConstraints.REMAINDER;
		c.weightx = 0.0;
		c.weighty = 0.0;

		Panel newFile = new Panel();
		newFile.setLayout(new GridLayout(1,1,2,2));

		newFile.add(new Button("New"));
		gridbag.setConstraints(newFile,  c);
		fileOptionsContainer.add(newFile);

		Panel ioPanel = editor.getFileIO().getGUI();
		gridbag.setConstraints(ioPanel,  c);
		fileOptionsContainer.add(ioPanel);

		Panel aboutPanel = new Panel();
		aboutPanel.setLayout(new GridLayout(1,1,2,2));

		aboutPanel.add(new Button("About"));

		c.fill = GridBagConstraints.BOTH;
		c.anchor = GridBagConstraints.CENTER;
		c.gridheight = GridBagConstraints.REMAINDER;
		c.weightx = 1.0;
		c.weighty = 1.0;

		gridbag.setConstraints(aboutPanel,  c);
		fileOptionsContainer.add(aboutPanel);

		fileOptions.add(fileOptionsContainer);

		optionsPanel.add("File",fileOptions);

		// Add the mode and options panels

		gridbag = new GridBagLayout();
		c = new GridBagConstraints();
		setLayout(gridbag);

		c.insets = new Insets(2,2,1,2);

		c.fill = GridBagConstraints.BOTH;
		c.anchor = GridBagConstraints.NORTH;
		c.gridwidth  = GridBagConstraints.REMAINDER;
		c.weightx = 0.0;
		c.weighty = 0.0;
		gridbag.setConstraints(modePanel,  c);
		add(modePanel);

		c.insets = new Insets(1,2,2,2);

		c.fill = GridBagConstraints.BOTH;
		c.anchor = GridBagConstraints.CENTER;
		c.gridheight = GridBagConstraints.REMAINDER;
		c.weightx = 1.0;
		c.weighty = 1.0;
		gridbag.setConstraints(optionsPanel,  c);
		add(optionsPanel);
	}

	// event dispatcher for various SwitchPanels. Events get sent here, from here
	// they are processed and UI layout is updated, and the events are forwarded to
	// UIPanel
	public void setUIEvent(int event) {
		UIPanel target = editor.getUI();
		switch(event)
		{
			case UIPanel.DRAW:
				((CardLayout)optionsPanel.getLayout()).show(optionsPanel,"Draw");
				editor.showUI();
				target.setUIMode(event);
				break;
			case UIPanel.GRID:
				((CardLayout)optionsPanel.getLayout()).show(optionsPanel,"Grid");
				editor.showUI();
				target.setUIMode(event);
				break;

			case UIPanel.CAMERA:
				((CardLayout)optionsPanel.getLayout()).show(optionsPanel,"Camera");
				editor.showUI();
				target.setUIMode(event);
				break;
			case UIPanel.FILE:
				((CardLayout)optionsPanel.getLayout()).show(optionsPanel,"File");
				editor.showMessages();
				target.setUIMode(event);
				break;
			case UIPanel.DRAW_FREEHAND:
				target.setDrawMode(event);
				break;
			case UIPanel.DRAW_LINE:
				target.setDrawMode(event);
				break;
			case UIPanel.DRAW_ERASER:
				target.setDrawMode(event);
				break;
			case UIPanel.GRID_MOVE:
				target.setGridMode(event);
				break;
			case UIPanel.GRID_PUSHROLL:
				target.setGridMode(event);
				break;
			case UIPanel.GRID_ROTATE:
				target.setGridMode(event);
				break;
			case UIPanel.GRID_BEND:
				target.setGridMode(event);
				break;
			case UIPanel.CAMERA_MOVE:
				target.setCameraMode(event);
				break;
			case UIPanel.CAMERA_PANTILT:
				target.setCameraMode(event);
				break;
			case UIPanel.CAMERA_ROTATE:
				target.setCameraMode(event);
				break;
			case UIPanel.CAMERA_DOLLYROLL:
				target.setCameraMode(event);
				break;
			case UIPanel.CAMERA_ZOOM:
				target.setCameraMode(event);
				break;
			default:
				break;
		}
	}

	public void notifyColorChange(ColorPicker picker) {
		UIPanel target = editor.getUI();
		if (picker == colorPicker)
		{
			if(target.isHistoryView())
				target.setCurrentHistoryColor(picker.getCurrentColor());
			else
				target.setCurrentDrawingColor(picker.getCurrentColor());
		}
	}

	public void notifyFOVChange(Camera cam) {
		Float fov = new Float(cam.getFov());
		cameraFOV.setText(fov.toString());
		repaint();
	}

	public void notifyGridChange(Grid grid) {
		Float size = new Float(grid.getSize());
		gridSize.setText(size.toString());
		Float step = new Float(grid.getStepSize());
		gridStepSize.setText(step.toString());
		gridStep.setState(grid.getStepping());
		gridSnap.setState(grid.getSnapping());
		repaint();
	}

	public boolean action(Event e, Object arg) {
		UIPanel target = editor.getUI();
		if (e.target instanceof Button) {
			String buttonText = (String)arg;

			if (target.getUIMode() == UIPanel.DRAW) {
				if (buttonText.equals("Undo")) {
					Scene scene = target.getScene();
					scene.undoCommand();
				}
				else if (buttonText.equals("Redo")) {
					Scene scene = target.getScene();
					scene.redoCommand();
				}
			}
			else if (target.getUIMode() == UIPanel.GRID) {
				if (buttonText.equals("Reset")) {
					target.resetGrid();
				}
			}
			else if (target.getUIMode() == UIPanel.FILE) {
				if (buttonText.equals("New")) {
					QuerySpawner query = new QuerySpawner();
					if (query.doQuery("","This will delete your drawing and start a new one.\nAre you sure?","Yes","No",null) == 1)
					{
						Scene scene = target.getScene();
						scene.reset();
//						FileIO io = editor.getFileIO();
						editor.notifySceneLoaded();
					}
				}
				else if (buttonText.equals("About")) {
					QuerySpawner query = new QuerySpawner();
					query.doQuery("","3d6b Editor v.1.0 09/01/2002\nDesign and programming by Kiia Kallio\nhttp://mlab.uiah.fi/projects/3D6B/","Ok",null,null);
				}
			}
		}
		else if (e.target instanceof Checkbox) {
			String checkboxText = ((Checkbox)e.target).getLabel();
			if (target.getUIMode() == UIPanel.DRAW) {
				if (checkboxText.equals("History")) {
					target.setHistoryView(((Checkbox)e.target).getState());
					if (target.isHistoryView())
						colorPicker.setCurrentColor(target.getCurrentHistoryColor());
					else
						colorPicker.setCurrentColor(target.getCurrentDrawingColor());
				}
			}
			else if (target.getUIMode() == UIPanel.GRID) {
				if (checkboxText.equals("Hide")) {
					target.setGridVisibility(!(((Checkbox)e.target).getState()));
				}
				else if (checkboxText.equals("Step:")) {
					target.setGridStepping((((Checkbox)e.target).getState()));
				}
				else if (checkboxText.equals("Snap")) {
					target.setGridSnapping((((Checkbox)e.target).getState()));
				}
			}
			else if (target.getUIMode() == UIPanel.CAMERA) {
				if (checkboxText.equals("Antialias")) {
					target.setAntialias(((Checkbox)e.target).getState());
				}
			}
		}
		else if (e.target instanceof TextField) {
			if (target.getUIMode() == UIPanel.GRID) {
				if (((TextField)e.target) == gridStepSize) {
					Float step;
					try {
						step = new Float(((TextField)e.target).getText());
					}
					catch (NumberFormatException exc) {
						step = new Float(0);
					}
					float stepValue = step.floatValue();

					if (stepValue < 0.01f)
						stepValue = 0.01f;
					else if (stepValue > 100f)
						stepValue = 100f;

					target.setGridStepSize(stepValue);

					step = new Float(stepValue);
					((TextField)e.target).setText(step.toString());
				}
				else if (((TextField)e.target) == gridSize) {
					Float size;
					try {
						size = new Float(((TextField)e.target).getText());
					}
					catch (NumberFormatException exc) {
						size = new Float(10);
					}
					float sizeValue = size.floatValue();

					if (sizeValue < 0.1f)
						sizeValue = 0.1f;
					else if (sizeValue > 100f)
						sizeValue = 100f;

					target.setGridSize(sizeValue);

					size = new Float(sizeValue);
					((TextField)e.target).setText(size.toString());
				}
			}
			else if (target.getUIMode() == UIPanel.CAMERA) {
				if (((TextField)e.target) == cameraFOV) {
					Float fov;
					try {
						fov = new Float(((TextField)e.target).getText());
					}
					catch (NumberFormatException exc) {
						fov = new Float(0);
					}
					float fovValue = fov.floatValue();
					if (fovValue < 5.0f)
						fovValue = 5.0f;
					else if (fovValue > 170.0f)
						fovValue = 170.0f;

					target.setFov(fovValue);

					fov = new Float(fovValue);
					((TextField)e.target).setText(fov.toString());
				}
			}
		}
		return true;
	}

	// Dispatch keyUp and keyDown events to UIPanel so that keyboard shortcuts
	// work. Events get dispatched only if the scene is visible and they are
	// not needed by some of the components of the editor.
	public boolean keyDown(Event e, int key) {
		UIPanel target = editor.getUI();
		if (target.getUIMode() != UIPanel.FILE &&
			e.target != gridStepSize &&
			e.target != gridSize &&
			e.target != cameraFOV) {
			return target.handleEvent(e);
		}
		return false;
	}

	public boolean keyUp(Event e, int key) {
		UIPanel target = editor.getUI();
		if (target.getUIMode() != UIPanel.FILE &&
			e.target != gridStepSize &&
			e.target != gridSize &&
			e.target != cameraFOV) {
			return target.handleEvent(e);
		}
		return false;
	}

}

interface QueryListener {
	public void notifyQuery(int result);
}


/*
 * This is a class for spawning query window and waiting for the input
 */
class QuerySpawner implements QueryListener {
	int result;

	QuerySpawner() {
	}

	public int doQuery(String title,String message,String text1,String text2,String text3) {
		Frame tf = new Frame();

		result = -1;
		new QueryPopup(tf,title,message,text1,text2,text3,this);

		while	(result == -1) {
			try {
				Thread.sleep(50);
			}
			catch (InterruptedException e) {
			}
		}

		tf.dispose();
		return result;
	}

	public void notifyQuery(int res) {
		result = res;
	}
}

/*
 * This is a popup dialog for querying ok/cancel questions.
 */
class QueryPopup extends Dialog {
	QueryListener resultListener;
	String buttonText1;
	String buttonText2;
	String buttonText3;

	QueryPopup(Frame parent,String title,String message,String text1,String text2,String text3,QueryListener listener) {
		super(parent,title,true);
		buttonText1 = text1;
		buttonText2 = text2;
		buttonText3 = text3;
		resultListener = listener;
		Panel p1 = new Panel();
		Panel p2 = new Panel();
		p2.setLayout(new FlowLayout(FlowLayout.CENTER,10,3));

		Vector strings = new Vector();
		String msg = message;
		int split = msg.indexOf("\n");
		while (split >= 0) {
			String begin = msg.substring(0,split);
			strings.addElement(begin);
			msg = msg.substring(split+1,msg.length());
			split = msg.indexOf("\n");
		}
		strings.addElement(msg);

		p1.setLayout(new GridLayout(strings.size(),1,3,3));

		for(int n = 0; n < strings.size(); n++) {
			Panel p = new Panel();
			p.setLayout(new BorderLayout(3,3));
			Label txt = new Label((String)strings.elementAt(n),Label.CENTER);
			p.add("Center",txt);
			p1.add(p);
		}

		if (buttonText1 != null)
			p2.add(new Button(buttonText1));
		if (buttonText2 != null)
			p2.add(new Button(buttonText2));
		if (buttonText3 != null)
			p2.add(new Button(buttonText3));
		setLayout(new BorderLayout(3,3));
		add("Center", p1);
		add("South", p2);
		setResizable(false);

/*		layout();
		Dimension dim = minimumSize();

		System.out.println("w: " + dim.width);
		System.out.println("h: " + dim.height);

		move(400 - dim.width/2,200 - dim.height/2);
*/
		resize(400,150);
		move(300,150);
		show();
	}

	public boolean action(Event e, Object arg) {
		if (e.target instanceof Button) {
			String buttonText = (String)arg;
			if (buttonText1 != null && buttonText.equals(buttonText1)) {
				resultListener.notifyQuery(1);
				dispose();
			}
			else if (buttonText2 != null && buttonText.equals(buttonText2)) {
				resultListener.notifyQuery(2);
				dispose();
			}
			else if (buttonText3 != null && buttonText.equals(buttonText3)) {
				resultListener.notifyQuery(3);
				dispose();
			}
		}
		return true;
	}

}

/*
 * MessagePanel is a panel for displaying scene info and messages
 */
class MessagesPanel extends Panel3D {
	Editor editor;
	TextArea historyArea;
	TextArea messageArea;

	MessagesPanel(Editor ed) {
		editor = ed;

		GridBagLayout gridbag = new GridBagLayout();
		GridBagConstraints c = new GridBagConstraints();
		setLayout(gridbag);
		Color bgColor = new Color((181+255)/2,(205+255)/2,(222+255)/2);
		setBackground(bgColor);

		c.insets = new Insets(4,4,4,4);

		c.fill = GridBagConstraints.NONE;
		c.anchor = GridBagConstraints.NORTHWEST;
		c.gridwidth  =  1;
		c.weightx = 0.0;
		c.weighty = 1.0;
		Label history = new Label("History:");
		gridbag.setConstraints(history,  c);
		add(history);

		c.fill = GridBagConstraints.BOTH;
		c.gridwidth = GridBagConstraints.REMAINDER;
            c.weightx  =  1.0;

		historyArea = new TextArea(10,80);
		historyArea.setEditable(false);
		gridbag.setConstraints(historyArea, c);
		add(historyArea);

		c.gridwidth  =  1;                // reset to the default
		c.weightx = 0.0;
		c.fill = GridBagConstraints.NONE;

		Label message = new Label("Message:");
		gridbag.setConstraints(message, c);
		add(message);

		c.fill = GridBagConstraints.BOTH;
		c.gridwidth = GridBagConstraints.REMAINDER;
            c.weightx  =  1.0;

		messageArea = new TextArea(10,80);
		messageArea.setEditable(true);
		gridbag.setConstraints(messageArea, c);
		add(messageArea);
	}

	public void setHistory(EditingHistory history) {
		String s = "";
		for (int n=0; n<history.getSize(); n++) {
			EditingHistoryEntry entry = history.getEntry(n);
			s += entry.getDescription() + "\n";
		}
		historyArea.setText(s);
	}

	public void setMessage(String text) {
		messageArea.setText(text);
	}

	public String getMessage() {
		return messageArea.getText();
	}
}

/*
 * Panel3D is a simple extension to AWT Panel. It draws the panel with
 * 3D outline.
 */
class Panel3D extends Panel {
	public void paint(Graphics g) {
		Rectangle r = bounds();

		g.setColor(getBackground());
		g.draw3DRect(0, 0, r.width-1, r.height-1, true);
	}
}


/*
 * UIEvent class is used by the SwitchPanel to store mappings between button
 * names and event ID's.
 */
class UIEvent {
	int event;
	String buttonText;
	Color buttonColor;

	UIEvent(String text, int event, Color col) {
		this.event = event;
		this.buttonText = text;
		this.buttonColor = col;
	}

	String getText() {
		return buttonText;
	}

	Color getColor() {
		return buttonColor;
	}

	int getEvent() {
		return event;
	}
}


/*
 * SwitchPanel is a class that combines a group of buttons into one panel.
 * Only one button at a time can be pressed down (in a same way as radio buttons).
 * When a button in SwitchPanel is clicked, the panel changes its state automatically
 * and makes a callback to UIControls group with given button event id number.
 */
class SwitchPanel extends Panel {
	Vector events = new Vector();
	UIControls target;

	SwitchPanel(UIControls target)
	{
		this.target = target;
	}

	void add(String text, int event, Color backgroundColor) {
		events.addElement(new UIEvent(text,event,backgroundColor));
		removeAll();
		setLayout(new CardLayout());

		int panelCount = events.size();

		ReferenceLabel tLabels[];
		Button tButtons[];

		tLabels = new ReferenceLabel[panelCount];
		tButtons = new Button[panelCount];

		int n,p;
		for (n=0; n<panelCount; n++) {
			Panel currentPanel = new Panel();

			GridBagLayout gridbag = new GridBagLayout();
			GridBagConstraints c = new GridBagConstraints();
			currentPanel.setLayout(gridbag);

			c.fill = GridBagConstraints.BOTH;
			c.anchor = GridBagConstraints.CENTER;
			c.gridwidth  = GridBagConstraints.REMAINDER;
			c.gridheight  = 1;
			c.weightx = 0.0;

			for (p=0; p<panelCount; p++) {
				if (p == 0)
					c.insets = new Insets(2,2,1,2);
				else if (p == panelCount-1) {
					c.gridheight = GridBagConstraints.REMAINDER;
					c.insets = new Insets(1,2,2,2);
				}
				else
					c.insets = new Insets(1,2,1,2);

				if (((UIEvent)events.elementAt(n)).getText().equals(((UIEvent)events.elementAt(p)).getText())) {
					DeepPanel tPanel = new DeepPanel();
					Color col = ((UIEvent)events.elementAt(p)).getColor();
					if (col != null)
						tPanel.setBackground(col);

					GridBagLayout gb = new GridBagLayout();
					GridBagConstraints gbc = new GridBagConstraints();
					tPanel.setLayout(gb);

					gbc.insets = new Insets(1,1,1,1);

					gbc.fill = GridBagConstraints.BOTH;
					gbc.anchor = GridBagConstraints.CENTER;
					gbc.gridwidth  = GridBagConstraints.REMAINDER;
					gbc.gridheight  = 1;
					gbc.weightx = 1.0;
					gbc.weighty = 1.0;

					tLabels[p] = new ReferenceLabel(((UIEvent)events.elementAt(p)).getText(),ReferenceLabel.CENTER);
					gb.setConstraints(tLabels[p],gbc);
					tPanel.add(tLabels[p]);

					gridbag.setConstraints(tPanel, c);
					currentPanel.add(tPanel);
				}
				else {
					Panel tPanel = new Panel();
					GridBagLayout gb = new GridBagLayout();
					GridBagConstraints gbc = new GridBagConstraints();
					tPanel.setLayout(gb);

					gbc.insets = new Insets(1,1,1,1);

					gbc.fill = GridBagConstraints.BOTH;
					gbc.anchor = GridBagConstraints.CENTER;
					gbc.gridwidth  = GridBagConstraints.REMAINDER;
					gbc.gridheight  = 1;
					gbc.weightx = 1.0;
					gbc.weighty = 1.0;

					tButtons[p] = new Button(((UIEvent)events.elementAt(p)).getText());
					gb.setConstraints(tButtons[p],gbc);
					tPanel.add(tButtons[p]);
					gridbag.setConstraints(tPanel, c);
					currentPanel.add(tPanel);
				}
			}
			add(((UIEvent)events.elementAt(n)).getText(), currentPanel);
		}
		for (n=0; n<panelCount; n++) {
			tLabels[n].setReferenceButton(tButtons[n]);
		}
	}	

	void setCurrent(String text) {
		((CardLayout)getLayout()).show(this,text);
		// without requestFocus the focus will stay in the hidden
		// button, and thus all keyboard events will go there and
		// disappear
		requestFocus();
	}

	public boolean action(Event e, Object arg) {
		if (e.target instanceof Button) {
			String buttonText = (String)arg;

			int panelCount = events.size();
			int n;
			for (n=0; n<panelCount; n++) {
				if (buttonText.equals(((UIEvent)events.elementAt(n)).getText())) {
					target.setUIEvent(((UIEvent)events.elementAt(n)).getEvent());
					setCurrent(buttonText);
					break;
				}
			}
		}
		return true;
	}

}


/*
 * ReferenceLabel is a component that is a label with reference to a button.
 * It uses the dimensions of the referenced button in size queries to have
 * identical layout with the button.
 */
class ReferenceLabel extends Label {
	Button referenceButton;
	boolean useRefButton;

	ReferenceLabel(String label, int align)
	{
		setText(label);
		setAlignment(align);
		useRefButton = false;
	}

	void setReferenceButton(Button b)
	{
		referenceButton = b;
		useRefButton = true;
	}

	public Dimension preferredSize()
	{
		if (useRefButton)
			return referenceButton.preferredSize();
		else
			return super.preferredSize();
	}

	public Dimension minimumSize()
	{
		if (useRefButton)
			return referenceButton.minimumSize();
		else
			return super.minimumSize();
	}
}

/*
 * DeepPanel is a component that is a Panel with 3D border.
 */
class DeepPanel extends Panel {
	DeepPanel() {}

	public void paint(Graphics g) {
		Rectangle r = bounds();

		g.setColor(getBackground());
		g.draw3DRect(0, 0, r.width-1, r.height-1, false);
	}
}


/*
 * ColorPicker is a class for implementing a RGB color picker with a
 * display of current color.
 */
class ColorPicker extends Panel {
	int width, height;
	int currentColor;
	Image imageGradient;
	Graphics gGradient;
	UIControls target;

	ColorPicker(UIControls target, int width, int height) {
		this.target = target;

		// width of the color picker area. Current color is
		// width x height, and border is one extra to every direction
		this.width = width;
		this.height = height;

		// init current color to black
		currentColor = 0;

		imageGradient = null;
		gGradient = null;
	}

	public int getCurrentColor()
	{
		return currentColor;
	}

	public void setCurrentColor(int color)
	{
		currentColor = color;
		makeGradient();
		repaint();
	}

	public Dimension preferredSize()
	{
		return new Dimension(width+2,2*height+2);
	}

	public Dimension minimumSize()
	{
		return new Dimension(width+2,2*height+2);
	}

	public boolean handleEvent(Event e) {
		Rectangle r = bounds();

		int xp = (r.width - (width + 2))/2;
		int yp = (r.height - (height*2 + 2))/2;

		switch (e.id) {
			case Event.MOUSE_DOWN:
			case Event.MOUSE_UP:
			case Event.MOUSE_DRAG:
				if (e.y>=yp+height+1 && e.y<=yp+2*height+1)
				{
					int tx = e.x-1-xp;
					if (tx<0)
						tx=0;
					else if (tx >= width)
						tx=width-1;

					int ty = e.y-height-1-yp;
					if (ty<0)
						ty=0;
					else if (ty >= height)
						ty=height-1;

					currentColor = calculateColor(currentColor,tx,ty);
					makeGradient();
					target.notifyColorChange(this);
					repaint();
				}
				return true;
			default:
				return false;
		}
	}

	private int calculateColor(int col, int x, int y)
	{
		int c = calculateShade(x);

		int yp = (3*y)/height;
		switch (yp) {
			// red
			case 0:
		 		c = c<<16;
				return (col & 0x00ffff) | c;
			// green
			case 1:
		 		c = c<<8;
				return (col & 0xff00ff) | c;
			// blue
			case 2:
				return (col & 0xffff00) | c;
			default:
				return 0;
		}
	}

	private int calculateShade(int x)
	{
		return (255*x)/(width-1);
	}

	private void makeGradient() {
		// create buffer for gradient
		if (imageGradient == null)
		{
			imageGradient = createImage(width, height);
			gGradient = imageGradient.getGraphics();
		}

		// draw the gradient
		int x,yp,c;
		yp = height/3;
		for (x=0; x<width; x++) {
			c = calculateShade(x);
			gGradient.setColor(new Color((currentColor & 0x00ffff) | (c<<16)));
			gGradient.drawLine(x,0,x,yp-1);
			gGradient.setColor(new Color((currentColor & 0xff00ff) | (c<<8)));
			gGradient.drawLine(x,yp,x,2*yp-1);
			gGradient.setColor(new Color((currentColor & 0xffff00) | c));
			gGradient.drawLine(x,2*yp,x,3*yp-1);
		}

		int rp = ((width-1)*((currentColor & 0xff0000)>>16))/255;
		int gp = ((width-1)*((currentColor & 0xff00)>>8))/255;
		int bp = ((width-1)*(currentColor & 0xff))/255;

		gGradient.setColor(new Color(0));
		gGradient.drawLine(rp,0,rp,yp-1);
		gGradient.drawLine(gp,yp,gp,2*yp-1);
		gGradient.drawLine(bp,2*yp,bp,3*yp-1);		
	}

	public void paint(Graphics g) {
		Rectangle r = bounds();

		int xp = (r.width - (width + 2))/2;
		int yp = (r.height - (height*2 + 2))/2;

		g.setColor(Color.lightGray);
		g.draw3DRect(xp, yp, width+1, 2*height+1, false);

		// blit the gradient
		if(imageGradient == null)
			makeGradient();
		g.drawImage(imageGradient, xp+1, yp+height+1, null);

		// draw current color
		g.setColor(new Color(currentColor));
		g.fillRect(xp+1,yp+1,width,height);
	}
}
