package v3d.editor;

import java.awt.*;
import java.io.*;

// 3d-line class
// - clip the line
// - draw the line
// - get the distance to a point
// - get the distance to a line segment
// - cut the line with a sphere
// - set the color of the line
// - get the color of the line
// - get the starting vertex number
// - get the ending vertex number
// - set the starting vertex number
// - set the ending vertex number
// - set the editor's id
// - serialize/deserialize
// - convert to VRML
class Line {
	public int startIdx;
	public int endIdx;
	public int color;
	public int editorId;
	public int nextEmptyIdx;
	public float zOffset;

	Line() {
		startIdx = 0;
		endIdx = 0;
		color = 0;
		editorId = 0;
		nextEmptyIdx = 0;
		zOffset = 0.0f;
	}

	Line(Line line) {
		this.copy(line);
	}

	public void copy(Line line) {
		this.startIdx = line.startIdx;
		this.endIdx = line.endIdx;
		this.color = line.color;
		this.editorId = line.editorId;
		this.nextEmptyIdx = line.nextEmptyIdx;
		this.zOffset = line.zOffset;
	}

	public void print() {
		System.out.println("start: " + this.startIdx +
					" end: " + this.endIdx +
					" color:" + this.color +
					" editor:" + this.editorId +
					" nextEmpty:" + this.nextEmptyIdx +
					" zOffset:" + this.zOffset);
	}

	// Line is defined with a line starting with word "l". The values on the line define
	// the index of the starting vertex of the line (one integer), the index of the ending
	// vertex of the line (one integer), colour of the line (three integers) and the index
	// to the editor entry of the line (one integer)

	// Creates a string of the line
	public String serialize() {
		String s = new String();
		s =	"l " +
			startIdx + " " +
			endIdx + " " +
			((color & 0xff0000) >> 16) + " " +
			((color & 0x00ff00) >> 8) + " " +
			(color & 0x0000ff) + " " +
			editorId + "\n";
		return s;
	}

	// Tries to read a line from the stream. Returns
	// true if the token was correct and data was read,
	// otherwise false.
	public boolean deserialize(StreamTokenizer st) throws IOException {
		if ("l".equals(st.sval)) {
			if (st.nextToken() == StreamTokenizer.TT_NUMBER)
				this.startIdx = (int)st.nval;
			if (st.nextToken() == StreamTokenizer.TT_NUMBER)
				this.endIdx = (int)st.nval;
			int r=0, g=0, b=0;
			if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
				r = (int)st.nval;
				if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
					g = (int)st.nval;
					if (st.nextToken() == StreamTokenizer.TT_NUMBER)
						b = (int)st.nval;
				}
			}
			this.color = (r << 16) | (g << 8) | b;

			if (st.nextToken() == StreamTokenizer.TT_NUMBER)
				this.editorId = (int)st.nval;

			while (st.ttype != StreamTokenizer.TT_EOL &&
				st.ttype != StreamTokenizer.TT_EOF)
				st.nextToken();

			this.zOffset = 0.0f;

			return true;
		}
		else {
			return false;
		}
	}

	public String getVRMLcolor() {
		String s = new String();
		s = (float)((color & 0xff0000)>>16)/255.0f + " " +
			(float)((color & 0x00ff00)>>8)/255.0f + " " +
			(float)(color & 0x00ff)/255.0f;
		return s;
	}

	public String getVRMLline() {
		String s = new String();
		s = startIdx + " " + endIdx + " -1";
		return s;
	}
}

// Line buffer is comparable to a vetex buffer, except that it holds a
// set of lines instead of a set of vertices.
class LineBuffer {
	Line lines[];
	int sortArray[];

	int bufferSize;
	int lineCount;
	int firstEmptyIdx;
	int lastEmptyIdx;
	int currentEditorId;
	int previousLine;
	int sortArraySize;

	int prevColor;

	UndoBuffer undo = null;

	LineBuffer(int size) {
		lines = new Line[size];
		for (int i=0; i<size; i++) {
			lines[i] = new Line();
		}
		bufferSize = size;
		sortArraySize = 0;
		clear();
	}

	public synchronized void clear() {
		for (int i=0; i<bufferSize; i++) {
			lines[i].nextEmptyIdx = i+1;
		}
		firstEmptyIdx = 0;
		lastEmptyIdx = bufferSize-1;
		lineCount = 0;
		previousLine = 0;
	}

	public synchronized void setCurrentEditor(int editorId) {
		currentEditorId = editorId;
	}

	public synchronized void resize(int size) {
		storeUndo(firstEmptyIdx);

		resizeDirect(size);

		storeRedo(firstEmptyIdx);
	}

	public void resizeDirect(int size) {
		if (size > bufferSize) {
			Line tl[];
			tl = new Line[bufferSize];
			for(int i=0; i<bufferSize; i++) {
				tl[i] = new Line(lines[i]);
			}

			this.lines = new Line[size];
			for(int i=0; i<bufferSize; i++) {
				lines[i] = new Line(tl[i]);
			}
			for(int i=bufferSize; i<size; i++) {
				lines[i] = new Line();
				lines[i].nextEmptyIdx = i+1;
			}
			lines[lastEmptyIdx].nextEmptyIdx = bufferSize;
			lastEmptyIdx = size - 1;
			bufferSize = size;
		}
		else if (size < bufferSize) {
			Line tl[];
			tl = new Line[bufferSize];
			for(int i=0; i<bufferSize; i++) {
				tl[i] = new Line(lines[i]);
			}
			this.lines = new Line[size];
			for(int i=0; i<size; i++) {
				lines[i] = new Line(tl[i]);
			}
			lastEmptyIdx = size - 1;
			bufferSize = size;
		}
	}

	// Duplicates the line buffer
	public synchronized LineBuffer duplicate() {
		LineBuffer lb = new LineBuffer(bufferSize);
		lb.copy(this);
		return lb;
	}

	// Copies the line buffer from another buffer
	public synchronized void copy(LineBuffer source) {
		if(bufferSize < source.bufferSize)
			resize(source.bufferSize);

		for(int i=0; i<source.bufferSize; i++) {
			lines[i].copy(source.lines[i]);
		}

		lineCount = source.lineCount;
		firstEmptyIdx = source.firstEmptyIdx;
		lastEmptyIdx = source.lastEmptyIdx;
		currentEditorId = source.currentEditorId;
		previousLine = source.previousLine;

		if(bufferSize > source.bufferSize) {
			for(int i=source.bufferSize; i<bufferSize; i++) {
				lines[i].nextEmptyIdx = i+1;
			}
			lines[lastEmptyIdx].nextEmptyIdx = source.bufferSize;
			lastEmptyIdx = bufferSize - 1;
		}
	}

	// Joins the buffer to another buffer, along with all the vertices.
	// Vertices may become duplicated. This is meant to be used for connecting
	// linebuffers for sorting after all transformations, projections and clipping
	// are done, so it really doesn't matter (except memory-wise)
	public synchronized void join(VertexBuffer sourceVertices, LineBuffer targetLines, VertexBuffer targetVertices) {
		int n=0, p=0;
		while(n<lineCount && p<bufferSize) {
			if(lines[p].nextEmptyIdx < 0) {
				int line = targetLines.line(targetVertices,
							sourceVertices.vertices[lines[p].startIdx].vec,
							sourceVertices.vertices[lines[p].endIdx].vec,
							lines[p].color);
				targetLines.lines[line].editorId = lines[p].editorId;
				targetLines.lines[line].zOffset = lines[p].zOffset;
				n++;
			}
			p++;
		}
	}

	public synchronized int line(VertexBuffer vertices, Vector3D start, Vector3D end, int color) {
		storeUndo(firstEmptyIdx);

		if(firstEmptyIdx == lastEmptyIdx) {
			resize(bufferSize + bufferSize/2 + 1);
		}
		int lineId = firstEmptyIdx;

		lines[lineId].startIdx = vertices.addVectorWithReference(start);
		lines[lineId].endIdx = vertices.addVectorWithReference(end);
		lines[lineId].color = color;
		lines[lineId].editorId = currentEditorId;

		firstEmptyIdx = lines[lineId].nextEmptyIdx;
		lines[lineId].nextEmptyIdx = -1;
		lineCount++;
		previousLine = lineId;

		storeRedo(lineId);

		return lineId;
	}

	public synchronized void line(VertexBuffer vertices, Line line) {
		directLine(line.startIdx, line.endIdx, line.color, line.editorId);
		vertices.addReference(line.startIdx);
		vertices.addReference(line.endIdx);
	}

	public synchronized void changeLineEnd(VertexBuffer vertices, int lineId, Vector3D endVec) {
		storeUndo(lineId);

		vertices.removeReference(lines[lineId].endIdx);

		lines[lineId].endIdx = vertices.addVectorWithReference(endVec);
		lines[lineId].editorId = currentEditorId;

		storeRedo(lineId);
	}

	public synchronized void changeLineStart(VertexBuffer vertices, int lineId, Vector3D startVec) {
		storeUndo(lineId);

 		vertices.removeReference(lines[lineId].startIdx);

		lines[lineId].startIdx = vertices.addVectorWithReference(startVec);
		lines[lineId].editorId = currentEditorId;

		storeRedo(lineId);
	}

	public synchronized void changeLineColor(int lineId, int color) {
		storeUndo(lineId);

		lines[lineId].color = color;
		lines[lineId].editorId = currentEditorId;

		storeRedo(lineId);
	}

	public synchronized void splitLine(VertexBuffer vertices, int lineId, Vector3D pos1, Vector3D pos2) {
		storeUndo(lineId);

		int pos1Idx = vertices.addVectorWithReference(pos1);
		int pos2Idx = vertices.addVectorWithReference(pos2);

		directLine(pos2Idx, lines[lineId].endIdx, lines[lineId].color, currentEditorId);

		lines[lineId].endIdx = pos1Idx;
		lines[lineId].editorId = currentEditorId;

		storeRedo(lineId);
	}

	public synchronized void removeLine(VertexBuffer vertices, int lineId) {
		storeUndo(lineId);

		vertices.removeReference(lines[lineId].startIdx);
		vertices.removeReference(lines[lineId].endIdx);

		lines[lineId].nextEmptyIdx = lines[lastEmptyIdx].nextEmptyIdx;
		lines[lastEmptyIdx].nextEmptyIdx = lineId;
		lastEmptyIdx = lineId;
		lineCount--;

		storeRedo(lineId);
	}

	public synchronized int directLine(int startIdx, int endIdx, int color, int editorId) {
		storeUndo(firstEmptyIdx);

		if(firstEmptyIdx == lastEmptyIdx) {
			resize(bufferSize + bufferSize/2 + 1);
		}
		int lineId = firstEmptyIdx;

		lines[lineId].startIdx = startIdx;
		lines[lineId].endIdx = endIdx;
		lines[lineId].color = color;
		lines[lineId].editorId = editorId;

		firstEmptyIdx = lines[lineId].nextEmptyIdx;
		lines[lineId].nextEmptyIdx = -1;
		lineCount++;
		previousLine = lineId;

		storeRedo(lineId);

		return lineId;
	}

	public synchronized int lineFromPreviousEnd(VertexBuffer vertices, Vector3D end, int color) {
		storeUndo(firstEmptyIdx);

		if(firstEmptyIdx == lastEmptyIdx) {
			resize(bufferSize + bufferSize/2 + 1);
		}
		int lineId = firstEmptyIdx;

		lines[lineId].startIdx = lines[previousLine].endIdx;
		vertices.addReference(lines[lineId].startIdx);
		lines[lineId].endIdx = vertices.addVectorWithReference(end);
		lines[lineId].color = color;
		lines[lineId].editorId = currentEditorId;

		firstEmptyIdx = lines[lineId].nextEmptyIdx;
		lines[lineId].nextEmptyIdx = -1;
		lineCount++;
		previousLine = lineId;

		storeRedo(lineId);

		return lineId;
	}

	public synchronized int lineFromPreviousStart(VertexBuffer vertices, Vector3D end, int color) {
		storeUndo(firstEmptyIdx);

		if(firstEmptyIdx == lastEmptyIdx) {
			resize(bufferSize + bufferSize/2 + 1);
		}
		int lineId = firstEmptyIdx;

		lines[lineId].startIdx = lines[previousLine].startIdx;
		vertices.addReference(lines[lineId].startIdx);
		lines[lineId].endIdx = vertices.addVectorWithReference(end);
		lines[lineId].color = color;
		lines[lineId].editorId = currentEditorId;

		firstEmptyIdx = lines[lineId].nextEmptyIdx;
		lines[lineId].nextEmptyIdx = -1;
		lineCount++;
		previousLine = lineId;

		storeRedo(lineId);

		return lineId;
	}

	public synchronized void sphereErase(VertexBuffer vertices, Vector3D spherePosition, float sphereRadius) {
		float radiusSq = sphereRadius*sphereRadius;

		int n=0, p=0;
		int startCount = lineCount;
		while(n<startCount && p<bufferSize) {
			if(lines[p].nextEmptyIdx < 0) {
				LineSphereIntersection intersect =
					new LineSphereIntersection(vertices.vertices[lines[p].startIdx].vec,
									vertices.vertices[lines[p].endIdx].vec,
									spherePosition, radiusSq);
				switch(intersect.getResult()) {
					// clip away the start of the line
					case 1:
						changeLineStart(vertices, p, intersect.getEndPoint());
						break;
					// clip away the end of the line
					case 2:
						changeLineEnd(vertices, p, intersect.getStartPoint());
						break;
					// clip away the middle of the line
					case 3:
						splitLine(vertices, p, intersect.getStartPoint(), intersect.getEndPoint());
						break;
					// erase the whole line
					case 4:
						removeLine(vertices, p);
						break;
				}
				n++;
			}
			p++;
		}
	}

	void storeUndo(int index) {
		if (undo != null) {
			LineBufferUndoCmd cmd = new LineBufferUndoCmd(this);
			storeState(cmd.state,index);
			undo.storeToQueue(cmd);
		}
	}

	void storeRedo(int index) {
		if (undo != null) {
			LineBufferRedoCmd cmd = new LineBufferRedoCmd(this);
			storeState(cmd.state,index);
			undo.storeToQueue(cmd);
		}
	}

	public void setUndoBuffer(UndoBuffer ub) {
		undo = ub;
	}

	// stores the state of the line buffer to LineBufferState
	void storeState(LineBufferState state, int index) {
		//System.out.println("LineBuffer.storeState()");
		state.index = index;
		state.line.copy(lines[index]);
		state.firstEmptyIdx = firstEmptyIdx;
		state.lastEmptyIdx = lastEmptyIdx;
		state.lastEmptyLine.copy(lines[lastEmptyIdx]);
		state.lineCount = lineCount;
		state.bufferSize = bufferSize;
		state.previousLine = previousLine;
	}

	// restores the state from LineBufferState
	public synchronized void restoreState(LineBufferState state) {
		//System.out.println("LineBuffer.restoreState()");
		if (state.bufferSize != bufferSize) {
			resizeDirect(state.bufferSize);
		}

		lines[state.index].copy(state.line);
		firstEmptyIdx = state.firstEmptyIdx;
		lastEmptyIdx = state.lastEmptyIdx;
		lines[state.lastEmptyIdx].copy(state.lastEmptyLine);
		lineCount = state.lineCount;
		previousLine = state.previousLine;
	}

	public synchronized void setZOffset(float zOffset) {
		int n=0, p=0;
		while(n<lineCount && p<bufferSize) {
			if(lines[p].nextEmptyIdx < 0) {
				lines[p].zOffset = zOffset;
				n++;
			}
			p++;
		}
	}

	public void zClip(VertexBuffer vertices, float minZ, float maxZ) {
		int n=0, p=0;
		int startCount = lineCount;
		while(n<startCount && p<bufferSize) {
			if(lines[p].nextEmptyIdx < 0) {
				float z1 = vertices.vertices[lines[p].startIdx].vec.z;
				float z2 = vertices.vertices[lines[p].endIdx].vec.z;

				if(z1 < minZ) {
					if(z2 < minZ) {
						vertices.removeReference(lines[p].startIdx);
						vertices.removeReference(lines[p].endIdx);
						lines[p].nextEmptyIdx = lines[lastEmptyIdx].nextEmptyIdx;
						lines[lastEmptyIdx].nextEmptyIdx = p;
						lastEmptyIdx = p;
						lineCount--;
					}
					else {
						Vector3D start = zClipLine(vertices.vertices[lines[p].startIdx].vec,
										vertices.vertices[lines[p].endIdx].vec,
										minZ);
						vertices.removeReference(lines[p].startIdx);
						lines[p].startIdx = vertices.addVectorWithReference(start);

						if (z2 > maxZ) {
							Vector3D end = zClipLine(start,
											vertices.vertices[lines[p].endIdx].vec,
											maxZ);
							vertices.removeReference(lines[p].endIdx);
							lines[p].endIdx = vertices.addVectorWithReference(end);
						}
					}
				}
				else if(z2 < minZ) {
					Vector3D end = zClipLine(vertices.vertices[lines[p].startIdx].vec,
									vertices.vertices[lines[p].endIdx].vec,
									minZ);
					vertices.removeReference(lines[p].endIdx);
					lines[p].endIdx = vertices.addVectorWithReference(end);

					if (z1 > maxZ) {
						Vector3D start = zClipLine(vertices.vertices[lines[p].startIdx].vec,
										end,
										maxZ);
						vertices.removeReference(lines[p].startIdx);
						lines[p].startIdx = vertices.addVectorWithReference(start);
					}
				}
				else if(z1 > maxZ) {
					if(z2 > maxZ) {
						vertices.removeReference(lines[p].startIdx);
						vertices.removeReference(lines[p].endIdx);
						lines[p].nextEmptyIdx = lines[lastEmptyIdx].nextEmptyIdx;
						lines[lastEmptyIdx].nextEmptyIdx = p;
						lastEmptyIdx = p;
						lineCount--;
					}
					else {
						Vector3D start = zClipLine(vertices.vertices[lines[p].startIdx].vec,
										vertices.vertices[lines[p].endIdx].vec,
										maxZ);
						vertices.removeReference(lines[p].startIdx);
						lines[p].startIdx = vertices.addVectorWithReference(start);
					}
				}
				else if(z2 > maxZ) {
					Vector3D end = zClipLine(vertices.vertices[lines[p].startIdx].vec,
									vertices.vertices[lines[p].endIdx].vec,
									maxZ);
					vertices.removeReference(lines[p].endIdx);
					lines[p].endIdx = vertices.addVectorWithReference(end);
				}
				n++;
			}
			p++;
		}
	}

	public void xClip(VertexBuffer vertices, float minX, float maxX) {
		int n=0, p=0;
		int startCount = lineCount;
		while(n<startCount && p<bufferSize) {
			if(lines[p].nextEmptyIdx < 0) {
				float x1 = vertices.vertices[lines[p].startIdx].vec.x;
				float x2 = vertices.vertices[lines[p].endIdx].vec.x;

				if(x1 < minX) {
					if(x2 < minX) {
						vertices.removeReference(lines[p].startIdx);
						vertices.removeReference(lines[p].endIdx);
						lines[lastEmptyIdx].nextEmptyIdx = p;
						lastEmptyIdx = p;
						lines[p].nextEmptyIdx = bufferSize+1;
						lineCount--;
					}
					else {
						Vector3D start = xClipLine(vertices.vertices[lines[p].startIdx].vec,
										vertices.vertices[lines[p].endIdx].vec,
										minX);
						vertices.removeReference(lines[p].startIdx);
						lines[p].startIdx = vertices.addVectorWithReference(start);

						if (x2 > maxX) {
							Vector3D end = xClipLine(start,
											vertices.vertices[lines[p].endIdx].vec,
											maxX);
							vertices.removeReference(lines[p].endIdx);
							lines[p].endIdx = vertices.addVectorWithReference(end);
						}
					}
				}
				else if(x2 < minX) {
					Vector3D end = xClipLine(vertices.vertices[lines[p].startIdx].vec,
									vertices.vertices[lines[p].endIdx].vec,
									minX);
					vertices.removeReference(lines[p].endIdx);
					lines[p].endIdx = vertices.addVectorWithReference(end);

					if (x1 > maxX) {
						Vector3D start = xClipLine(vertices.vertices[lines[p].startIdx].vec,
										end,
										maxX);
						vertices.removeReference(lines[p].startIdx);
						lines[p].startIdx = vertices.addVectorWithReference(start);
					}
				}
				else if(x1 > maxX) {
					if(x2 > maxX) {
						vertices.removeReference(lines[p].startIdx);
						vertices.removeReference(lines[p].endIdx);
						lines[p].nextEmptyIdx = lines[lastEmptyIdx].nextEmptyIdx;
						lines[lastEmptyIdx].nextEmptyIdx = p;
						lastEmptyIdx = p;
						lineCount--;
					}
					else {
						Vector3D start = xClipLine(vertices.vertices[lines[p].startIdx].vec,
										vertices.vertices[lines[p].endIdx].vec,
										maxX);
						vertices.removeReference(lines[p].startIdx);
						lines[p].startIdx = vertices.addVectorWithReference(start);
					}
				}
				else if(x2 > maxX) {
					Vector3D end = xClipLine(vertices.vertices[lines[p].startIdx].vec,
									vertices.vertices[lines[p].endIdx].vec,
									maxX);
					vertices.removeReference(lines[p].endIdx);
					lines[p].endIdx = vertices.addVectorWithReference(end);
				}
				n++;
			}
			p++;
		}
	}

	public void yClip(VertexBuffer vertices, float minY, float maxY) {
		int n=0, p=0;
		int startCount = lineCount;
		while(n<startCount && p<bufferSize) {
			if(lines[p].nextEmptyIdx < 0) {
				float y1 = vertices.vertices[lines[p].startIdx].vec.y;
				float y2 = vertices.vertices[lines[p].endIdx].vec.y;

				if(y1 < minY) {
					if(y2 < minY) {
						vertices.removeReference(lines[p].startIdx);
						vertices.removeReference(lines[p].endIdx);
						lines[p].nextEmptyIdx = lines[lastEmptyIdx].nextEmptyIdx;
						lines[lastEmptyIdx].nextEmptyIdx = p;
						lastEmptyIdx = p;
						lineCount--;
					}
					else {
						Vector3D start = yClipLine(vertices.vertices[lines[p].startIdx].vec,
										vertices.vertices[lines[p].endIdx].vec,
										minY);
						vertices.removeReference(lines[p].startIdx);
						lines[p].startIdx = vertices.addVectorWithReference(start);

						if (y2 > maxY) {
							Vector3D end = yClipLine(start,
											vertices.vertices[lines[p].endIdx].vec,
											maxY);
							vertices.removeReference(lines[p].endIdx);
							lines[p].endIdx = vertices.addVectorWithReference(end);
						}
					}
				}
				else if(y2 < minY) {
					Vector3D end = yClipLine(vertices.vertices[lines[p].startIdx].vec,
									vertices.vertices[lines[p].endIdx].vec,
									minY);
					vertices.removeReference(lines[p].endIdx);
					lines[p].endIdx = vertices.addVectorWithReference(end);

					if (y1 > maxY) {
						Vector3D start = yClipLine(vertices.vertices[lines[p].startIdx].vec,
										end,
										maxY);
						vertices.removeReference(lines[p].startIdx);
						lines[p].startIdx = vertices.addVectorWithReference(start);
					}
				}
				else if(y1 > maxY) {
					if(y2 > maxY) {
						vertices.removeReference(lines[p].startIdx);
						vertices.removeReference(lines[p].endIdx);
						lines[p].nextEmptyIdx = lines[lastEmptyIdx].nextEmptyIdx;
						lines[lastEmptyIdx].nextEmptyIdx = p;
						lastEmptyIdx = p;
						lineCount--;
					}
					else {
						Vector3D start = yClipLine(vertices.vertices[lines[p].startIdx].vec,
										vertices.vertices[lines[p].endIdx].vec,
										maxY);
						vertices.removeReference(lines[p].startIdx);
						lines[p].startIdx = vertices.addVectorWithReference(start);
					}
				}
				else if(y2 > maxY) {
					Vector3D end = yClipLine(vertices.vertices[lines[p].startIdx].vec,
									vertices.vertices[lines[p].endIdx].vec,
									maxY);
					vertices.removeReference(lines[p].endIdx);
					lines[p].endIdx = vertices.addVectorWithReference(end);
				}
				n++;
			}
			p++;
		}
	}

	Vector3D zClipLine(Vector3D start, Vector3D end, float zPlane)
	{
		Vector3D vec = new Vector3D();

/*
				x1,z1
			     /
			    /
			   /
			--+--- z0
		^	 /
		|	/
		z	x2,z2
		 x->

		z = ax + b
		
		z1 = ax1 + b
		z2 = ax2 + b

		z1 - ax1 = b
		z2 - ax2 = b
		z1 - ax1 = z2 - ax2

		ax1 - ax2 = z1 - z2
		a (x1 - x2) = z1 - z2
		a = (z1 - z2) / (x1 - x2)

		b = z1 - ax1

		ax = z - b;
		x = (z - b)/a
		x = (z - z1 + ax1)/a
		x = (z - z1)/a + x1
		x = (x1 - x2)(z - z1)/(z1 - z2) + x1

*/
		float zComp = (zPlane - start.z)/(start.z - end.z);

		vec.x = (start.x - end.x) * zComp + start.x;
		vec.y = (start.y - end.y) * zComp + start.y;
		vec.z = zPlane;

		return vec;
	}

	Vector3D xClipLine(Vector3D start, Vector3D end, float xPlane)
	{
		Vector3D vec = new Vector3D();

		float xComp = (xPlane - start.x)/(start.x - end.x);

		vec.x = xPlane;
		vec.y = (start.y - end.y) * xComp + start.y;
		vec.z = (start.z - end.z) * xComp + start.z;

		return vec;
	}

	Vector3D yClipLine(Vector3D start, Vector3D end, float yPlane)
	{
		Vector3D vec = new Vector3D();

		float yComp = (yPlane - start.y)/(start.y - end.y);

		vec.x = (start.x - end.x) * yComp + start.x;
		vec.y = yPlane;
		vec.z = (start.z - end.z) * yComp + start.z;

		return vec;
	}

	public synchronized void drawSorted(VertexBuffer vertices, Graphics g, int sortBucketCount, float minZ, float maxZ) {
		int n, p, zi, i;
		float bucketStep, midZ;

		if (sortArraySize < sortBucketCount) {
			sortArray = new int[sortBucketCount];
			sortArraySize = sortBucketCount;
		}

		bucketStep = (maxZ - minZ) / (float)(sortBucketCount - 1);

		// init buckets
		for(n=0; n<sortBucketCount; n++) {
			sortArray[n] = -1;
		}

		// sort to buckets
		n=0;
		p=0;
		while(n<lineCount && p<bufferSize) {
			if(lines[p].nextEmptyIdx < 0) {
				midZ = (vertices.vertices[lines[p].startIdx].vec.z + 
						vertices.vertices[lines[p].endIdx].vec.z + lines[p].zOffset)/2.0f;
				zi = (int)((midZ - minZ)/bucketStep);

				// use nextEmptyIdx for storing the list within the bucket
				lines[p].nextEmptyIdx = sortArray[zi];
				sortArray[zi] = p;

				n++;
			}
			p++;
		}

		// init prevColor to non-existing color
		prevColor = -1;
		// draw from buckets
		for(n=sortBucketCount-1; n>=0; n--) {
			i = sortArray[n];
			while(i>=0) {
				drawLine(g,
						vertices.vertices[lines[i].startIdx].vec,
						vertices.vertices[lines[i].endIdx].vec,
						lines[i].color);
				zi = i;
				i = lines[zi].nextEmptyIdx;
				// reset the line to be recognized as existing line
				lines[zi].nextEmptyIdx = -1;
			}
		}
	}

	public synchronized void drawSortedHistory(VertexBuffer vertices, Graphics g, int sortBucketCount, float minZ, float maxZ, EditingHistory history) {
		int n, p, zi, i;
		float bucketStep, midZ;

		if (sortArraySize < sortBucketCount) {
			sortArray = new int[sortBucketCount];
			sortArraySize = sortBucketCount;
		}

		bucketStep = (maxZ - minZ) / (float)(sortBucketCount - 1);

		// init buckets
		for(n=0; n<sortBucketCount; n++) {
			sortArray[n] = -1;
		}

		// sort to buckets
		n=0;
		p=0;
		while(n<lineCount && p<bufferSize) {
			if(lines[p].nextEmptyIdx < 0) {
				midZ = (vertices.vertices[lines[p].startIdx].vec.z + 
						vertices.vertices[lines[p].endIdx].vec.z + lines[p].zOffset)/2.0f;
				zi = (int)((midZ - minZ)/bucketStep);

				// use nextEmptyIdx for storing the list within the bucket
				lines[p].nextEmptyIdx = sortArray[zi];
				sortArray[zi] = p;

				n++;
			}
			p++;
		}

		// init prevColor to non-existing color
		prevColor = -1;
		// draw from buckets
		for(n=sortBucketCount-1; n>=0; n--) {
			i = sortArray[n];
			while(i>=0) {
				int eid = lines[i].editorId;
				int col;
				if (eid >= 0)
					col = history.getEntry(eid).getColor();
				else
					col = lines[i].color;

				drawLine(g,
						vertices.vertices[lines[i].startIdx].vec,
						vertices.vertices[lines[i].endIdx].vec,
						col);
				zi = i;
				i = lines[zi].nextEmptyIdx;
				// reset the line to be recognized as existing line
				lines[zi].nextEmptyIdx = -1;
			}
		}
	}

	public synchronized void drawUnsorted(VertexBuffer vertices, Graphics g) {
		int n=0, p=0;
		while(n<lineCount && p<bufferSize) {
			if(lines[p].nextEmptyIdx < 0) {
				drawLine(g,
						vertices.vertices[lines[p].startIdx].vec,
						vertices.vertices[lines[p].endIdx].vec,
						lines[p].color);
				drawLine(g,
						new Vector3D(0.0f,0.0f,0.0f),
						new Vector3D(10.0f,10.0f,0.0f),
						0xff0000);
				n++;
			}
			p++;
		}
	}

	void drawLine(Graphics g, Vector3D start, Vector3D end, int color) {
		// speed up the color switching by storing the previous color
		if (prevColor != color)
		{
			g.setColor(new Color(color));
			prevColor = color;
		}
		g.drawLine((int)start.x, (int)start.y, (int)end.x, (int)end.y);
	}

	synchronized void print(VertexBuffer sourceVertices) {
		int n=0, p=0;
		while(n<lineCount && p<bufferSize) {
			if(lines[p].nextEmptyIdx < 0) {
				//System.out.println("Start:");
				sourceVertices.vertices[lines[p].startIdx].vec.print();
				//System.out.println("End:");
				sourceVertices.vertices[lines[p].endIdx].vec.print();
				n++;
			}
			p++;
		}
	}

	public synchronized void print() {
		System.out.println("bufferSize: " + bufferSize);
		System.out.println("lineCount: " + lineCount);
		System.out.println("firstEmptyIdx: " + firstEmptyIdx);
		System.out.println("lastEmptyIdx: " + lastEmptyIdx);
		int n=0, p=0;
		while(n<lineCount && p<bufferSize) {
			if(lines[p].nextEmptyIdx < 0) {
				lines[p].print();
				n++;
			}
			p++;
		}
	}

	public synchronized void makeFog(VertexBuffer vertices, float minZ, float maxZ) {
		int n=0, p=0;
		float zMul = 255.0f/(maxZ - minZ);
		while(n<lineCount && p<bufferSize) {
			if(lines[p].nextEmptyIdx < 0) {
				float midZ = (vertices.vertices[lines[p].startIdx].vec.z + 
						vertices.vertices[lines[p].endIdx].vec.z)/2.0f;
				int shade = (int)((maxZ - midZ)*zMul);
				int invShade = 255 - shade;

				int r = (lines[p].color & 0xff0000) >> 16;
				int g = (lines[p].color & 0x00ff00) >> 8;
				int b = (lines[p].color & 0x0000ff);

				int i = r * shade + 128;
                    	r = ((( i + (i >> 8) ) >> 8) + invShade) << 16;
				
				i = g * shade + 128;
                    	g = ((( i + (i >> 8) ) >> 8) + invShade) << 8;

				i = b * shade + 128;
                    	b = ((( i + (i >> 8) ) >> 8) + invShade);

				lines[p].color = r | g | b;

				n++;
			}
			p++;
		}
	}

	public synchronized void compress(VertexBuffer vertices) {
		// first compress the vertex buffer
		int[] vertexIdxArray = vertices.compress();

		// change the index references to match compressed values
		int n=0, p=0;
		while(n<lineCount && p<bufferSize) {
			if(lines[p].nextEmptyIdx < 0) {
				lines[p].startIdx = vertexIdxArray[lines[p].startIdx];
				lines[p].endIdx = vertexIdxArray[lines[p].endIdx];
				n++;
			}
			p++;
		}

		// compress the line buffer
		n = 0; p = 0;
		while (p<bufferSize) {
			if(lines[p].nextEmptyIdx < 0) {
				if (p > n) {
					lines[n].copy(lines[p]);
					// tag the line to be empty
					lines[p].nextEmptyIdx = 0;
				}
				n++;
			}
			p++;
		}

		firstEmptyIdx = n;
		lastEmptyIdx = bufferSize;
		for (p = firstEmptyIdx; p < lastEmptyIdx; p++) {
			lines[p].nextEmptyIdx = p+1;
		}
	}

	// Serializes the line buffer to a string. Note that the buffer
	// should have been compressed (along with the vertex buffer) prior to
	// serializing.
	public synchronized String serialize() {
		String s = new String();
		int n=0, p=0;
		while(n<lineCount && p<bufferSize) {
			if(lines[p].nextEmptyIdx < 0) {
				s += lines[p].serialize();
				n++;
			}
			p++;
		}
		s += "\n";
		return s;
	}

	public synchronized String getVRMLcolors(int indent) {
		String sIndent = new String();
		for (int i=0; i<indent; i++)
			sIndent += "\t";

		String s;
		s = sIndent + "color Color {\n"+
			sIndent + "\tcolor [\n";

		int n=0, p=0;
		while(n<lineCount && p<bufferSize) {
			if(lines[p].nextEmptyIdx < 0) {
				s += sIndent + "\t\t" + lines[p].getVRMLcolor() + ",\n";
				n++;
			}
			p++;
		}

		s += sIndent + "\t\t]\n" +
			sIndent + "}\n";

		return s;
	}

	public synchronized String getHistoryVRMLcolors(int indent, EditingHistory history) {
		String sIndent = new String();
		for (int i=0; i<indent; i++)
			sIndent += "\t";

		String s;
		s = sIndent + "color Color {\n"+
			sIndent + "\tcolor [\n";

		int n=0, p=0;
		while(n<lineCount && p<bufferSize) {
			if(lines[p].nextEmptyIdx < 0) {
				s += sIndent + "\t\t" + history.getEntry(lines[p].editorId).getVRMLcolor() + ",\n";
				n++;
			}
			p++;
		}

		s += sIndent + "\t\t]\n" +
			sIndent + "}\n";

		return s;
	}

	public synchronized String getVRMLlines(int indent) {
		String sIndent = new String();
		for (int i=0; i<indent; i++)
			sIndent += "\t";

		String s;
		s = sIndent + "coordIndex [\n";

		int n=0, p=0;
		while(n<lineCount && p<bufferSize) {
			if(lines[p].nextEmptyIdx < 0) {
				s += sIndent + "\t" + lines[p].getVRMLline() + "\n";
				n++;
			}
			p++;
		}

		s += sIndent + "\t]\n";

		return s;
	}
}


// 3d-line set class
// - draw the line set (from a given camera)
//    - project the lines
//    - clip the lines
//    - sort the lines
// - add a line
// - remove a line
// - find lines based on 3d-point and radius
// - find lines based on ray
// - cut parts from lines
// - command history buffer
// - editing information buffer
// - serialize/deserialize line set
// - convert to VRML (with editing information parameters)
// - contains vertex buffer
// - repack lines to the vertex buffer
class LineSet {
	public LineBuffer originalLines;
	public LineBuffer tempLines;
	public VertexBuffer originalBuffer;
	public VertexBuffer transformedBuffer;
	public VertexBuffer projectedBuffer;
	UndoBuffer undo = null;

	LineSet(int size) {
		originalLines = new LineBuffer(size);
		tempLines = new LineBuffer(size);

		originalBuffer = new VertexBuffer(size*2);
		transformedBuffer = new VertexBuffer(size*2);
		projectedBuffer = new VertexBuffer(size*2);
	}

	public synchronized void setCurrentEditor(int editor) {
		originalLines.setCurrentEditor(editor);
		tempLines.setCurrentEditor(editor);
	}

	public synchronized int line(Vector3D start, Vector3D end, int color) {
		return originalLines.line(originalBuffer, start, end, color);
	}

	public synchronized int lineFromPreviousEnd(Vector3D end, int color) {
		return originalLines.lineFromPreviousEnd(originalBuffer, end, color);
	}

	public synchronized int lineFromPreviousStart(Vector3D end, int color) {
		return originalLines.lineFromPreviousStart(originalBuffer, end, color);
	}

	public synchronized void changeLineEnd(int lineId, Vector3D end) {
		originalLines.changeLineEnd(originalBuffer, lineId, end);
	}

	public synchronized void changeLineStart(int lineId, Vector3D start) {
		originalLines.changeLineStart(originalBuffer, lineId, start);
	}

	public synchronized void changeLineColor(int lineId, int color) {
		originalLines.changeLineColor(lineId, color);
	}

	public synchronized void erase(Vector3D pos, float radius) {
		originalLines.sphereErase(originalBuffer, pos, radius);
	}

	public synchronized void setUndoBuffer(UndoBuffer ub) {
		undo = ub;
		originalLines.setUndoBuffer(ub);
		originalBuffer.setUndoBuffer(ub);
	}

 	public synchronized void prepareForDraw(Camera cam, LineBuffer targetLines, VertexBuffer targetVertices) {
		originalBuffer.transform(cam.viewMatrix,transformedBuffer);
		tempLines.copy(originalLines);
		tempLines.zClip(transformedBuffer,cam.minZ,cam.maxZ);
		transformedBuffer.project(cam,projectedBuffer);
		tempLines.xClip(projectedBuffer,cam.viewXOffset,cam.viewXOffset+cam.viewWidth);
		tempLines.yClip(projectedBuffer,cam.viewYOffset,cam.viewYOffset+cam.viewHeight);

		tempLines.makeFog(projectedBuffer,cam.getProjectedZPlane(cam.minZ),cam.getProjectedZPlane(cam.maxZ));

		tempLines.join(projectedBuffer, targetLines, targetVertices);
	}

	// clears the lineset
	public synchronized void clear() {
		originalLines.clear();
		originalBuffer.clear();
	}

	// Serializes the lineset to a string
	public synchronized String serialize() {
		String s = new String();
		originalLines.compress(originalBuffer);
		s += originalBuffer.serialize();
		s += originalLines.serialize();

		return s;
	}

	// Tries to read a lineset item from the stream. Returns
	// true if the token was correct and new item was added,
	// otherwise false. Items can be either vertices or lines.
	public synchronized boolean deserialize(StreamTokenizer st) throws IOException {
		if (originalBuffer.deserialize(st)) {
			return true;
		}
		else {
			Line line = new Line();
			if (line.deserialize(st)) {
				originalLines.line(originalBuffer,line);
				return true;
			}
			else {
				return false;
			}
		}
	}

	public synchronized String getVRML(int indent) {
		String sIndent = new String();
		for (int i=0; i<indent; i++)
			sIndent += "\t";

		String s;
		s = sIndent + "geometry IndexedLineSet {\n" +
			sIndent + "\tcolorPerVertex FALSE\n";

		s += originalLines.getVRMLcolors(indent+1);
		s += originalBuffer.getVRML(indent+1);
		s += originalLines.getVRMLlines(indent+1);

		s += sIndent + "}\n";

		return s;
	}

	public synchronized String getHistoryVRML(int indent, EditingHistory history) {
		String sIndent = new String();
		for (int i=0; i<indent; i++)
			sIndent += "\t";

		String s;
		s = sIndent + "geometry IndexedLineSet {\n" +
			sIndent + "\tcolorPerVertex FALSE\n";

		s += originalLines.getHistoryVRMLcolors(indent+1,history);
		s += originalBuffer.getVRML(indent+1);
		s += originalLines.getVRMLlines(indent+1);

		s += sIndent + "}\n";

		return s;

	}
}

class LineSphereIntersection {
	int result;
	Vector3D startResult;
	Vector3D endResult;

	// calculates intersections of a line with a sphere
	// returns 0 if there was no intersections, 1 if there was one intersection and the
	// start of the line was inside the sphere, 2 if there was one intersection and the
	// end of the line was inside the sphere, 3 if there was two intersections, and 4
	// if the line was totally inside the sphere.
	public LineSphereIntersection(Vector3D lineStart, Vector3D lineEnd, Vector3D spherePosition, float radiusSq) {
		Vector3D l, p, e;
		startResult = new Vector3D(0,0,0);
		endResult = new Vector3D(0,0,0);

		p = spherePosition.subtract(lineStart);
		l = lineEnd.subtract(lineStart);
		e = spherePosition.subtract(lineEnd);

		float pSq = p.lengthSq();
		float eSq = e.lengthSq();

		result = 0;
		// both ends of the line are inside the sphere
		if (pSq < radiusSq) {
			if (eSq < radiusSq) {
				result = 4;
				return;
			}
			else {
				result = 1;
			}
		}
		else if (eSq < radiusSq) {
			result = 2;
		}

		Vector3D minPoint = new Vector3D();
		float prod, lenSq = 0;

		prod = p.dotProd(l);
		lenSq = l.lengthSq();

		// (p.l)*l/|l|^2
		minPoint = l.multiply(prod/lenSq);
		minPoint.addToThis(lineStart);

		Vector3D diff = spherePosition.subtract(minPoint);
		float diffLenSq = diff.lengthSq();

		if (result == 0 && prod > 0 && prod < lenSq && diffLenSq < radiusSq) {
			result = 3;
		}

		if (result == 0)
			return;

		float aLen;
		Vector3D a;

		// |diff|^2 + |a|^2 = r^2
		// |a|^2 = r^2 - diff^2
		// |a| = sqrt(r^2 - diff^2);
		// a = l0 * |a|
		// a = l/|l| * |a|
		aLen = (float)Math.sqrt(radiusSq - diffLenSq);
		a = l.multiply(aLen/(float)Math.sqrt(lenSq));

		// result == 1 || result == 3
		if (result != 2) {
			endResult.copy(minPoint);
			endResult.addToThis(a);
		}

		// result == 2 || result == 3
		if (result >= 2) {
			startResult.copy(minPoint);
			startResult.subtractFromThis(a);
		}
	}

	public int getResult() {
		return result;
	}

	public Vector3D getStartPoint() {
		return startResult;
	}

	public Vector3D getEndPoint() {
		return endResult;
	}
}
