package v3d.editor;

import java.awt.*;
import java.io.*;
import java.util.Vector;

class FileFormatException extends Exception {
	public FileFormatException(String s) {
		super(s);
	}
}

// scene class
// holds together the various aspects of the scene
// - deserialize scene
// - save scene
//	- save editing information
//	- create image
//	- create VRML files
//	- save text description
// 	- serialize scene
class Scene {
	public Grid grid;
	public LineSet lineset;
	public Camera cam;
	public EditingHistory history;
	LineBuffer displayLines;
	VertexBuffer displayVertices;
	public UndoBuffer undo = null;

	boolean gridVisible;

	Scene(int width, int height) {

		grid = new Grid(20,15.0f,0xc0c0c0);
		gridVisible = true;

		lineset = new LineSet(1000);
		// set the current editor id
		lineset.setCurrentEditor(0);

		history = new EditingHistory();

		cam = new Camera();
		cam.setBase(new Vector3D(0.0f,300.0f,0.0f),
				new Vector3D(0.0f,0.0f,0.0f),
				new Vector3D(1.0f,0.0f,0.0f));
		cam.setView((float)width,(float)height,60.0f, 10.0f, 2500.0f);
		cam.setViewOffset(0,0);

		displayLines = new LineBuffer(500);
		displayVertices = new VertexBuffer(1000);

		undo = new UndoBuffer(100);

		reset();
	}

	public synchronized void reset() {
		undo.clear();
		lineset.clear();
		history.clear();

		grid = new Grid(20,15.0f,0xc0c0c0);

		lineset.setCurrentEditor(0);

		cam.setBase(new Vector3D(0.0f,300.0f,0.0f),
				new Vector3D(0.0f,0.0f,0.0f),
				new Vector3D(1.0f,0.0f,0.0f));
		cam.setFov(60.0f);
		cam.setViewOffset(0,0);

		enableUndo();
	}

	public synchronized void showGrid() {
		gridVisible = true;
	}

	public synchronized void hideGrid() {
		gridVisible = false;
	}

	public synchronized void render(Graphics g) {
		displayLines.clear();
		displayVertices.clear();
		if (gridVisible)
			grid.prepareForDraw(cam, displayLines, displayVertices);
		lineset.prepareForDraw(cam, displayLines, displayVertices);
		displayLines.drawSorted(displayVertices,g,1000,
						cam.getProjectedZPlane(cam.minZ),
						cam.getProjectedZPlane(cam.maxZ));
//		displayLines.drawUnsorted(displayVertices,g);
	}

	public synchronized void renderHistory(Graphics g) {
		displayLines.clear();
		displayVertices.clear();
		if (gridVisible)
			grid.prepareForDraw(cam, displayLines, displayVertices);
		lineset.prepareForDraw(cam, displayLines, displayVertices);
		displayLines.drawSortedHistory(displayVertices,g,1000,
						cam.getProjectedZPlane(cam.minZ),
						cam.getProjectedZPlane(cam.maxZ),
						history);
	}

	public void setEditor(String editorInfo) {
		int col = history.getNextColor();
  		history.addEntry(new EditingHistoryEntry(col, editorInfo));
		int id = history.getSize() - 1;
		lineset.setCurrentEditor(id);
//		System.out.println("editor: "+id);
	}

	public synchronized void startCommand() {
		undo.openQueue();
	}

	public synchronized void endCommand() {
		undo.closeQueue();
	}

	public synchronized void undoCommand() {
		undo.undo();
	}

	public synchronized void redoCommand() {
		undo.redo();
	}

	public synchronized void enableUndo() {
		undo.clear();
		lineset.setUndoBuffer(undo);
		grid.setUndoBuffer(undo);
//		System.out.println("Undo enabled");
	}

	public synchronized void disableUndo() {
		lineset.setUndoBuffer(null);
		grid.setUndoBuffer(null);
//		System.out.println("Undo disabled");
	}

	// Deserializes the scene from input stream
	public synchronized void deserialize(InputStream is) throws IOException, FileFormatException {
		reset();
		disableUndo();
		StreamTokenizer st = new StreamTokenizer(is);
		st.eolIsSignificant(true);
		st.commentChar('#');
		st.quoteChar('"');
	scan:
		while (true) {
			switch (st.nextToken()) {
				case StreamTokenizer.TT_EOL:
					break;
				case StreamTokenizer.TT_WORD:
					if (cam.deserialize(st)) {
					}
					else if (grid.deserialize(st)) {
					}
					else if (lineset.deserialize(st)) {
					}
					else if (history.deserialize(st)) {
					}
					else {
						while (st.nextToken() != StreamTokenizer.TT_EOL
							&& st.ttype != StreamTokenizer.TT_EOF);
					}
					break;
				default:
					break scan;
			}
		}
		is.close();
		enableUndo();
		if (st.ttype != StreamTokenizer.TT_EOF)
			throw new FileFormatException(st.toString());
	}

	// Serializes the scene to a string
	public synchronized String serialize() {
		disableUndo();

		String s = new String();
		s += "# 3d6b file\n\n";
		s += "# Camera\n";
		s += cam.serialize();
		s += "\n# Grid\n";
		s += grid.serialize();
		s += "\n# Lineset\n";
		s += lineset.serialize();
		s += "\n# History\n";
		s += history.serialize();

		enableUndo();
		return s;
	}

	public synchronized String getVRML() {
		String s = new String();

		s += "#VRML V2.0 utf8\n" +
			"Background {\n" +
			"\tskyColor [1.0 1.0 1.0]\n" +
			"}\n";
		s += cam.getVRML();
		s +=	"Shape {\n" +
			"\tappearance Appearance {\n" +
			"\t\tmaterial Material {\n" +
			"\t\t}\n" +
			"\t}\n";
		s += lineset.getVRML(1);
		s += "}\n";

		return s;
	}

	public synchronized String getHistoryVRML() {
		String s = new String();

		s += "#VRML V2.0 utf8\n" +
			"Background {\n" +
			"\tskyColor [1.0 1.0 1.0]\n" +
			"}\n";
		s += cam.getVRML();
		s +=	"Shape {\n" +
			"\tappearance Appearance {\n" +
			"\t\tmaterial Material {\n" +
			"\t\t}\n" +
			"\t}\n";
		s += lineset.getHistoryVRML(1,history);
		s += "}\n";

		return s;
	}
}


// editing history class
// - store editor's information & editing session time -> get an id
// - request information based on an id
// - remove (unused) id's
// - serialize/deserialize
class EditingHistoryEntry {
	int color;
	String description;

	EditingHistoryEntry() {}

	EditingHistoryEntry(int color, String description) {
		this.color = color;
		this.description = description;
	}

	public void setColor(int color) {
		this.color = color;
	}

	public void setDescription(String description) {
		this.description = description;
	}

	public int getColor() {
		return color;
	}

	public String getDescription() {
		return description;
	}

	// Creates a string of the history entry
	public String serialize() {
		String s = new String();
		s =	"history " +
			((color & 0xff0000) >> 16) + " " +
			((color & 0x00ff00) >> 8) + " " +
			(color & 0x0000ff) + " " +
			"\"" + description + "\"\n";
		return s;
	}

	// Tries to read a history entry from the stream. Returns
	// true if the token was correct and values were read,
	// otherwise false.
	public boolean deserialize(StreamTokenizer st) throws IOException {
		if ("history".equals(st.sval)) {
			int r = 0, g = 0, b = 0;
			if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
				r = (int)st.nval;
				if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
					g = (int)st.nval;
					if (st.nextToken() == StreamTokenizer.TT_NUMBER)
						b = (int)st.nval;
				}
			}
			this.color = (r << 16) | (g << 8) | b;
		
			if (st.nextToken() == '"')
				this.description = st.sval;

			while (st.ttype != StreamTokenizer.TT_EOL &&
				st.ttype != StreamTokenizer.TT_EOF)
				st.nextToken();

			return true;
		}
		else {
			return false;
		}
	}

	public String getVRMLcolor() {
		String s = new String();
		s = (float)((color & 0xff0000)>>16)/255.0f + " " +
			(float)((color & 0x00ff00)>>8)/255.0f + " " +
			(float)(color & 0x00ff)/255.0f;
		return s;
	}
}

class EditingHistory {
	Vector history;
	int colorTable[];
	int colorCount;

	EditingHistory() {
		history = new Vector();
		colorTable = new int[256];
		colorCount = 256;

		// fill the array with proper palette, not random numbers!
		for (int i=0; i<256; i++) {
			int r = (int)(Math.random()*255);
			int g = (int)(Math.random()*255);
			int b = (int)(Math.random()*255);
			colorTable[i] = (r<<16)|(g<<8)|b;
		}
	}

	public void clear() {
		history.removeAllElements();
	}

	public void addEntry(EditingHistoryEntry entry) {
		history.addElement(entry);
	}

	public String serialize() {
		String s = new String();
		int i;
		for (i=0; i<history.size(); i++) {
			s += getEntry(i).serialize();
		}
		s += "\n";
		return s;
	}

	// Tries to read a history entry from the stream. Returns
	// true if the token was correct and new entry was added,
	// otherwise false.
	public boolean deserialize(StreamTokenizer st) throws IOException {
		EditingHistoryEntry entry = new EditingHistoryEntry();
		if (entry.deserialize(st)) {
			addEntry(entry);
			return true;
		}
			else {
			return false;
		}
	}

	public EditingHistoryEntry getEntry(int index) {
		return (EditingHistoryEntry)history.elementAt(index);
	}

	public int getSize() {
		return history.size();
	}

	public int getNextColor() {
		int size = history.size();
		if (size >= colorCount - 1) {
			int r = (int)(Math.random()*255);
			int g = (int)(Math.random()*255);
			int b = (int)(Math.random()*255);
			return (r<<16)|(g<<8)|b;
		}
		else {
			return colorTable[size + 1];
		}
	}

	public void setCurrentColor(int col) {
		int size = history.size();
		if (size > 0) {
			getEntry(size - 1).setColor(col);
		}
	}

	public int getCurrentColor() {
		int size = history.size();
		if (size > 0) {
			return getEntry(size - 1).getColor();
		}
		else {
			return colorTable[0];
		}
	}
}
